#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA @UnusedWildImport

import os
import unittest

import numpy as np

from obspy import UTCDateTime, read
from obspy.io.ah.core import _is_ah, _read_ah, _write_ah1, _read_ah1
from obspy.core.util import NamedTemporaryFile


class CoreTestCase(unittest.TestCase):
    """
    AH (Ad Hoc) file test suite.
    """
    def setUp(self):
        # Directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data')

    def test_is_ah(self):
        """
        Testing AH file format.
        """
        # AH v1
        testfile = os.path.join(self.path, 'TSG', 'BRV.TSG.DS.lE21.resp')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'TSG', 'BRV.TSG.KSM.sE12.resp')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah1.f')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah1.c')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah1.t')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'hrv.lh.zne')
        self.assertTrue(_is_ah(testfile))

        # AH v2
        testfile = os.path.join(self.path, 'ah2.f')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah2.f-e')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah2.c')
        self.assertTrue(_is_ah(testfile))
        testfile = os.path.join(self.path, 'ah2.t')
        self.assertTrue(_is_ah(testfile))

        # non AH files
        testfile = os.path.join(self.path, 'TSG', 'BRV.TSG.DS.lE21.asc')
        self.assertFalse(_is_ah(testfile))
        testfile = os.path.join(self.path, 'TSG', 'BRV.TSG.KSM.sE12.asc')
        self.assertFalse(_is_ah(testfile))
        testfile = os.path.join(self.path, 'TSG', 'Readme_TSG_response.txt')
        self.assertFalse(_is_ah(testfile))

    def test_read(self):
        """
        Testing reading AH file format using read() function.
        """
        # AH v1
        testfile = os.path.join(self.path, 'hrv.lh.zne')
        st = read(testfile)
        self.assertEqual(len(st), 3)
        testfile = os.path.join(self.path, 'ah1.f')
        st = read(testfile)
        self.assertEqual(len(st), 4)
        # not supported data types (vector, complex, tensor)
        testfile = os.path.join(self.path, 'ah1.c')
        self.assertRaises(NotImplementedError, _read_ah, testfile)
        testfile = os.path.join(self.path, 'ah1.t')
        self.assertRaises(NotImplementedError, _read_ah, testfile)

        # AH v2
        # float
        testfile = os.path.join(self.path, 'ah2.f')
        st = read(testfile)
        self.assertEqual(len(st), 4)

    def test_read_ah(self):
        """
        Testing reading AH file format using _read_ah() function.
        """
        # AH v1
        testfile = os.path.join(self.path, 'ah1.f')
        st = _read_ah(testfile)
        self.assertEqual(len(st), 4)
        tr = st[0]
        ah = tr.stats.ah
        # station
        self.assertEqual(ah.version, '1.0')
        self.assertEqual(ah.station.code, 'RSCP')
        self.assertEqual(ah.station.channel, 'IPZ')
        self.assertEqual(ah.station.type, 'null')
        self.assertAlmostEqual(ah.station.latitude, 35.599899, 6)
        self.assertAlmostEqual(ah.station.longitude, -85.568802, 6)
        self.assertEqual(ah.station.elevation, 481.0)
        self.assertAlmostEqual(ah.station.gain, 64200.121094, 6)
        self.assertEqual(len(ah.station.poles), 24)
        self.assertEqual(len(ah.station.zeros), 7)
        # event
        self.assertEqual(ah.event.latitude, 0.0)
        self.assertEqual(ah.event.longitude, 0.0)
        self.assertEqual(ah.event.depth, 0.0)
        self.assertEqual(ah.event.origin_time, None)
        self.assertEqual(ah.event.comment, 'null')
        # record
        self.assertEqual(ah.record.type, 1)
        self.assertEqual(ah.record.ndata, 720)
        self.assertEqual(tr.stats.npts, 720)
        self.assertEqual(len(tr), 720)
        self.assertEqual(tr.data.dtype, np.float64)
        self.assertAlmostEqual(ah.record.delta, 0.25, 6)
        self.assertAlmostEqual(tr.stats.delta, 0.25, 6)
        self.assertAlmostEqual(ah.record.max_amplitude, 0.0, 6)
        dt = UTCDateTime(1984, 4, 20, 6, 42, 0, 120000)
        self.assertEqual(ah.record.start_time, dt)
        self.assertEqual(tr.stats.starttime, dt)
        self.assertEqual(ah.record.comment, 'null')
        self.assertEqual(ah.record.log, 'gdsn_tape;demeaned;')
        # data
        np.testing.assert_array_almost_equal(tr.data[:4], np.array([
            -731.41247559, -724.41247559, -622.41247559, -470.4125061]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            -1421.41247559, 118.58750153, 88.58750153, -982.41247559]))

        # not supported data types (vector, complex, tensor)
        testfile = os.path.join(self.path, 'ah1.c')
        self.assertRaises(NotImplementedError, _read_ah, testfile)
        testfile = os.path.join(self.path, 'ah1.t')
        self.assertRaises(NotImplementedError, _read_ah, testfile)

        # AH v2
        testfile = os.path.join(self.path, 'ah2.f')
        st = _read_ah(testfile)
        self.assertEqual(len(st), 4)
        tr = st[0]
        ah = tr.stats.ah
        self.assertEqual(ah.version, '2.0')
        # station
        self.assertEqual(ah.station.code, 'RSCP')
        self.assertEqual(ah.station.channel, 'IPZ')
        self.assertEqual(ah.station.type, 'null')
        self.assertAlmostEqual(ah.station.latitude, 35.599899, 6)
        self.assertAlmostEqual(ah.station.longitude, -85.568802, 6)
        self.assertEqual(ah.station.elevation, 481.0)
        self.assertAlmostEqual(ah.station.gain, 64200.121094, 6)
        self.assertEqual(len(ah.station.poles), 24)
        self.assertEqual(len(ah.station.zeros), 7)
        # event
        self.assertEqual(ah.event.latitude, 0.0)
        self.assertEqual(ah.event.longitude, 0.0)
        self.assertEqual(ah.event.depth, 0.0)
        self.assertEqual(ah.event.origin_time, None)
        self.assertEqual(ah.event.comment, 'null')
        # record
        self.assertEqual(ah.record.type, 1)
        self.assertEqual(ah.record.ndata, 720)
        self.assertEqual(tr.stats.npts, 720)
        self.assertEqual(len(tr), 720)
        self.assertEqual(tr.data.dtype, np.float64)
        self.assertAlmostEqual(ah.record.delta, 0.25, 6)
        self.assertAlmostEqual(tr.stats.delta, 0.25, 6)
        self.assertAlmostEqual(ah.record.max_amplitude, 0.0, 6)
        dt = UTCDateTime(1984, 4, 20, 6, 42, 0, 120000)
        self.assertEqual(ah.record.start_time, dt)
        self.assertEqual(tr.stats.starttime, dt)
        self.assertEqual(ah.record.comment, 'null')
        self.assertEqual(ah.record.log, 'gdsn_tape;demeaned;')
        # data
        np.testing.assert_array_almost_equal(tr.data[:4], np.array([
            -731.41247559, -724.41247559, -622.41247559, -470.4125061]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            -1421.41247559, 118.58750153, 88.58750153, -982.41247559]))

        # not supported data types (vector, complex, tensor)
        testfile = os.path.join(self.path, 'ah2.t')
        self.assertRaises(NotImplementedError, _read_ah, testfile)

    def test_tsg(self):
        """
        Test reading AH v1 files of the STsR-TSG System at Borovoye.

        .. seealso:: https://www.ldeo.columbia.edu/res/pi/Monitoring/Data/
        """
        # 1 - BRV.TSG.DS.lE21
        testfile = os.path.join(self.path, 'TSG', 'BRV.TSG.DS.lE21.resp')
        st = _read_ah(testfile)
        self.assertEqual(len(st), 1)
        tr = st[0]
        ah = tr.stats.ah
        self.assertEqual(ah.version, '1.0')
        # station
        self.assertEqual(ah.station.code, 'BRVK')
        self.assertEqual(ah.station.channel, 'lE21')
        self.assertEqual(ah.station.type, 'TSG-DS')
        self.assertAlmostEqual(ah.station.latitude, 53.058060, 6)
        self.assertAlmostEqual(ah.station.longitude, 70.282799, 6)
        self.assertEqual(ah.station.elevation, 300.0)
        self.assertAlmostEqual(ah.station.gain, 0.05, 6)
        self.assertAlmostEqual(ah.station.normalization, 40.009960, 6)
        self.assertAlmostEqual(ah.station.longitude, 70.282799, 6)
        # calibration
        self.assertEqual(len(ah.station.poles), 7)
        self.assertAlmostEqual(ah.station.poles[0],
                               complex(-1.342653e-01, 1.168836e-01), 6)
        self.assertAlmostEqual(ah.station.poles[1],
                               complex(-1.342653e-01, -1.168836e-01), 6)
        self.assertEqual(len(ah.station.zeros), 4)
        self.assertAlmostEqual(ah.station.zeros[0], complex(0.0, 0.0), 6)
        self.assertAlmostEqual(ah.station.zeros[1], complex(0.0, 0.0), 6)
        self.assertAlmostEqual(ah.station.zeros[2], complex(0.0, 0.0), 6)
        self.assertAlmostEqual(ah.station.zeros[3], complex(0.0, 0.0), 6)
        # event
        self.assertAlmostEqual(ah.event.latitude, 49.833000, 6)
        self.assertAlmostEqual(ah.event.longitude, 78.807999, 6)
        self.assertEqual(ah.event.depth, 0.5)
        self.assertEqual(ah.event.origin_time, UTCDateTime(1988, 2, 8, 15, 23))
        self.assertEqual(ah.event.comment, 'Calibration_for_hg_TSG')
        # record
        self.assertEqual(ah.record.type, 1)
        self.assertEqual(ah.record.ndata, 225)
        self.assertEqual(tr.stats.npts, 225)
        self.assertEqual(len(tr), 225)
        self.assertEqual(tr.data.dtype, np.float64)
        self.assertAlmostEqual(ah.record.delta, 0.312, 6)
        self.assertAlmostEqual(tr.stats.delta, 0.312, 6)
        self.assertAlmostEqual(ah.record.max_amplitude, 785.805786, 6)
        dt = UTCDateTime(1988, 2, 8, 15, 24, 50.136002)
        self.assertEqual(ah.record.start_time, dt)
        self.assertEqual(tr.stats.starttime, dt)
        self.assertEqual(ah.record.abscissa_min, 0.0)
        self.assertEqual(ah.record.comment, 'DS response in counts/nm;')
        self.assertEqual(ah.record.log,
                         'brv2ah: ahtedit;demeaned;modhead;modhead;ahtedit;')
        # extras
        self.assertEqual(len(ah.extras), 21)
        self.assertEqual(ah.extras[0], 0.0)
        self.assertAlmostEqual(ah.extras[1], 0.1, 6)
        self.assertAlmostEqual(ah.extras[2], 0.1, 6)
        self.assertEqual(ah.extras[3], 0.0)
        # data
        np.testing.assert_array_almost_equal(tr.data[:24], np.array([
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            -1.19425595, -1.19425595, -1.19425595, -1.19425595,
            52.8057518, 175.80580139, 322.80578613, 463.80578613]))
        np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
            1.80574405, 2.80574393, 3.80574393, 3.80574393]))

    def test_write_ah1(self):
        """
        Testing writing AH1 file format using _write_ah1() function.
        """
        # AH v1
        testfile = os.path.join(self.path, 'st.ah')
        stream_orig = _read_ah(testfile)

        with NamedTemporaryFile() as tf:
            tmpfile = tf.name + '.AH'
            # write testfile
            _write_ah1(stream_orig, tmpfile)
            # read again
            st = _read_ah1(tmpfile)
            self.assertEqual(len(st), 1)
            tr = st[0]
            ah = tr.stats.ah
            stats = tr.stats
            # stream header
            self.assertEqual(stats.network, '')
            self.assertEqual(stats.station, 'ALE')
            self.assertEqual(stats.location, '')
            self.assertEqual(stats.channel, 'VHZ')
            starttime = UTCDateTime(1994, 6, 9, 0, 40, 45)
            endtime = UTCDateTime(1994, 6, 12, 8, 55, 4, 724522)
            self.assertEqual(stats.starttime, starttime)
            self.assertEqual(stats.endtime, endtime)
            self.assertAlmostEqual(stats.sampling_rate, 0.100000, 6)
            self.assertAlmostEqual(stats.delta, 9.999990, 6)
            self.assertEqual(stats.npts, 28887)
            self.assertEqual(len(tr), 28887)
            self.assertEqual(stats.calib, 1.0)

            # station
            self.assertEqual(ah.version, '1.0')
            self.assertEqual(ah.station.code, 'ALE')
            self.assertEqual(ah.station.channel, 'VHZ')
            self.assertEqual(ah.station.type, 'Global S')
            self.assertEqual(ah.station.latitude, 82.50330352783203)
            self.assertEqual(ah.station.longitude, -62.349998474121094)
            self.assertEqual(ah.station.elevation, 60.0)
            self.assertEqual(ah.station.gain, 265302864.0)
            self.assertEqual(len(ah.station.poles), 13)
            self.assertEqual(len(ah.station.zeros), 6)
            # event
            self.assertEqual(ah.event.latitude, -13.872200012207031)
            self.assertEqual(ah.event.longitude, -67.51249694824219)
            self.assertEqual(ah.event.depth, 640000.0)
            origintime = UTCDateTime(1994, 6, 9, 0, 33, 16)
            self.assertEqual(ah.event.origin_time, origintime)
            self.assertEqual(ah.event.comment, 'null')
            # record
            self.assertEqual(ah.record.type, 1)
            self.assertEqual(ah.record.ndata, 28887)
            self.assertEqual(tr.data.dtype, np.float64)
            self.assertAlmostEqual(ah.record.delta, 9.999990, 6)
            self.assertEqual(ah.record.max_amplitude, 9.265750885009766)
            rstarttime = UTCDateTime(1994, 6, 9, 0, 40, 45)
            self.assertEqual(ah.record.start_time, rstarttime)
            comment = 'Comp azm=0.0,inc=-90.0; Disp (m);'
            self.assertEqual(ah.record.comment, comment)
            self.assertEqual(ah.record.log, 'null')
            # data
            np.testing.assert_array_almost_equal(tr.data[:4], np.array([
                -236., -242., -252., -262.]))
            np.testing.assert_array_almost_equal(tr.data[-4:], np.array([
                101., 106., 107., 104.]))


def suite():
    return unittest.makeSuite(CoreTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import json

from .default import Default


def get_dump_kwargs(minify=True, no_nulls=True, **kwargs):
    """
    Return dict of kwargs for :py:func:`json.dump` or
    :py:func:`json.dumps`.

    :param bool minify: Use no spaces between separators (True)
    :param bool no_nulls: Omit null values and empty sequences/mappings (True)

    """
    if minify:
        kwargs["separators"] = (',', ':')
    kwargs["default"] = Default(omit_nulls=no_nulls)
    return kwargs


def _write_json(obj, filename, omit_nulls=False, pretty_print=True,
                **kwargs):
    """
    Write object to a file in JSON format

    :type obj: :mod:`~obspy.core.event` class object
    :param obj: The ObsPy Event-type object to write.
    :type filename: str or file
    :param filename: Filename to write or open file-like object.
    :type omit_nulls: bool
    :param omit_nulls: Don't include empty-valued attributes
    :type pretty_print: bool
    :param pretty_print: Indent for readability

    .. note::
        This function is registered via the
        :meth:`~obspy.core.event.Catalog.write` method of an ObsPy
        :class:`~obspy.core.event.Catalog` object, but is also valid for any
        obspy event-type object (or any serializable python object that
        contains an obspy event-type object)

    """
    try:
        # Open filehandler or use an existing file like object.
        if not hasattr(filename, "write"):
            file_opened = True
            fh = open(filename, "wt")
        else:
            file_opened = False
            fh = filename

        default = Default(omit_nulls=omit_nulls)
        if pretty_print:
            kwargs.setdefault('indent', 2)
        json_string = str(json.dumps(obj, default=default, **kwargs))
        fh.write(json_string)
    finally:
        # Close if a file has been opened by this function.
        if file_opened is True:
            fh.close()



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import os
import unittest
import warnings

from obspy.core.event import ResourceIdentifier, read_events
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.util.base import NamedTemporaryFile
from obspy.io.pde.mchedr import _read_mchedr
from obspy.io.quakeml.core import _read_quakeml, _write_quakeml

# lxml < 2.3 seems not to ship with RelaxNG schema parser and namespace support
IS_RECENT_LXML = False
try:
    from lxml.etree import __version__
    version = float(__version__.rsplit('.', 1)[0])
    if version >= 2.3:
        IS_RECENT_LXML = True
except Exception:
    pass


class MchedrTestCase(unittest.TestCase):
    """
    Test suite for obspy.mchedr
    """

    catalog = None

    def setUp(self):
        # directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data')
        filename = os.path.join(self.path, 'mchedr.dat')
        # read the mchedr file once for all
        if self.__class__.catalog is None:
            self.__class__.catalog = _read_mchedr(filename)

    def test_catalog(self):
        self.assertEqual(len(self.catalog), 1)
        self.assertEqual(
            str(self.catalog),
            '''1 Event(s) in Catalog:
2012-01-01T05:27:55.980000Z | +31.456, +138.072 | 6.2 Mb'''
        )

    def test_event(self):
        """
        Tests Event object.
        """
        event = self.catalog[0]
        self.assertEqual(
            event.resource_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/event/20120101052755.98'))
        # enums
        self.assertEqual(event.event_type, None)
        self.assertEqual(event.event_type_certainty, None)
        # comments
        self.assertEqual(len(event.comments), 1)
        c = event.comments
        self.assertEqual(c[0].text, 'MW 6.8 (WCMT), 6.8 (UCMT), 6.8 (GCMT). \
Felt (V) at Chiba; (IV) at Fussa, Kawasaki, Saitama, Tokyo, \
Yokohama and Yokosuka; (III) at Ebina, Zama and Zushi; (II) \
at Misawa and Narita, Honshu. Recorded (4 JMA) in Chiba, Fukushima, \
Gumma, Ibaraki, Kanagawa, Miyagi, Saitama, Tochigi and Tokyo.')
        # event descriptions
        self.assertEqual(len(event.event_descriptions), 2)
        d = event.event_descriptions
        self.assertEqual(d[0].text, 'SOUTHEAST OF HONSHU, JAPAN')
        self.assertEqual(d[0].type, 'region name')
        self.assertEqual(d[1].text, '211')
        self.assertEqual(d[1].type, 'Flinn-Engdahl region')
        # creation info
        self.assertEqual(event.creation_info, None)

    def test_origin(self):
        """
        Tests Origin object.
        """
        self.assertEqual(len(self.catalog[0].origins), 4)
        origin = self.catalog[0].origins[0]
        self.assertEqual(
            origin.resource_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/origin/20120101052755.98'))
        self.assertEqual(origin.origin_type, 'hypocenter')
        self.assertEqual(
            origin.time,
            UTCDateTime(2012, 1, 1, 5, 27, 55, 980000))
        self.assertEqual(origin.latitude, 31.456)
        self.assertAlmostEqual(
            origin.latitude_errors.uncertainty, 0.0155, places=3)
        self.assertEqual(origin.longitude, 138.072)
        self.assertAlmostEqual(
            origin.longitude_errors.uncertainty, 0.0173, places=3)
        self.assertEqual(origin.depth, 365300.0)
        self.assertEqual(origin.depth_errors.uncertainty, 2700.0)
        self.assertEqual(origin.depth_type, 'from location')
        self.assertEqual(origin.method_id, None)
        self.assertEqual(origin.time_fixed, None)
        self.assertEqual(origin.epicenter_fixed, None)
        self.assertEqual(
            origin.earth_model_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/earthmodel/ak135'))
        self.assertEqual(origin.evaluation_mode, None)
        self.assertEqual(origin.evaluation_status, None)
        self.assertEqual(origin.origin_type, 'hypocenter')
        # composite times
        self.assertEqual(len(origin.composite_times), 0)
        # quality
        self.assertEqual(origin.quality.used_station_count, 628)
        self.assertEqual(origin.quality.standard_error, 0.84)
        self.assertEqual(origin.quality.azimuthal_gap, 10.8)
        self.assertEqual(origin.quality.maximum_distance, 29.1)
        self.assertEqual(origin.quality.minimum_distance, 2.22)
        self.assertEqual(origin.quality.associated_phase_count, 52)
        self.assertEqual(origin.quality.associated_station_count, 628)
        self.assertEqual(origin.quality.depth_phase_count, 0)
        self.assertEqual(origin.quality.secondary_azimuthal_gap, None)
        self.assertEqual(origin.quality.ground_truth_level, None)
        self.assertEqual(origin.quality.median_distance, None)
        # comments
        self.assertEqual(len(origin.comments), 0)
        # creation info
        self.assertEqual(origin.creation_info.author, None)
        self.assertEqual(origin.creation_info.agency_id, 'USGS-NEIC')
        self.assertEqual(origin.creation_info.author_uri, None)
        self.assertEqual(origin.creation_info.agency_uri, None)
        self.assertEqual(origin.creation_info.creation_time, None)
        self.assertEqual(origin.creation_info.version, None)
        # origin uncertainty
        u = origin.origin_uncertainty
        self.assertEqual(u.preferred_description, 'confidence ellipsoid')
        self.assertEqual(u.horizontal_uncertainty, None)
        self.assertEqual(u.min_horizontal_uncertainty, None)
        self.assertEqual(u.max_horizontal_uncertainty, None)
        self.assertEqual(u.azimuth_max_horizontal_uncertainty, None)
        # confidence ellipsoid
        c = u.confidence_ellipsoid
        self.assertEqual(c.semi_intermediate_axis_length, 2750.0)
        # c.major_axis_rotation is computed during file reading:
        self.assertAlmostEqual(c.major_axis_rotation, 170.5, places=3)
        self.assertEqual(c.major_axis_plunge, 76.06)
        self.assertEqual(c.semi_minor_axis_length, 2210.0)
        self.assertEqual(c.semi_major_axis_length, 4220.0)
        self.assertEqual(c.major_axis_azimuth, 292.79)

    def test_magnitude(self):
        """
        Tests Magnitude object.
        """
        self.assertEqual(len(self.catalog[0].magnitudes), 3)
        mag = self.catalog[0].magnitudes[0]
        self.assertEqual(
            mag.resource_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/magnitude/20120101052755.98/mb'))
        self.assertEqual(mag.mag, 6.2)
        self.assertEqual(mag.mag_errors.uncertainty, None)
        self.assertEqual(mag.magnitude_type, 'Mb')
        self.assertEqual(mag.station_count, 294)
        self.assertEqual(mag.evaluation_status, None)
        # comments
        self.assertEqual(len(mag.comments), 0)
        # creation info
        self.assertEqual(mag.creation_info.author, None)
        self.assertEqual(mag.creation_info.agency_id, 'USGS-NEIC')
        self.assertEqual(mag.creation_info.author_uri, None)
        self.assertEqual(mag.creation_info.agency_uri, None)
        self.assertEqual(mag.creation_info.creation_time, None)
        self.assertEqual(mag.creation_info.version, None)

    def test_stationmagnitude(self):
        """
        Tests StationMagnitude object.
        """
        self.assertEqual(len(self.catalog[0].station_magnitudes), 19)
        mag = self.catalog[0].station_magnitudes[0]
        self.assertEqual(mag.mag, 6.6)
        self.assertEqual(mag.mag_errors.uncertainty, None)
        self.assertEqual(mag.station_magnitude_type, 'Mb')
        self.assertEqual(mag.waveform_id.station_code, 'MDJ')
        self.assertEqual(mag.creation_info, None)

    def test_amplitude(self):
        """
        Tests Amplitude object.
        """
        self.assertEqual(len(self.catalog[0].station_magnitudes), 19)
        amp = self.catalog[0].amplitudes[0]
        self.assertAlmostEqual(amp.generic_amplitude, 3.94502e-06)
        self.assertEqual(amp.type, 'AB')
        self.assertEqual(amp.period, 1.3)
        self.assertEqual(amp.magnitude_hint, 'Mb')
        self.assertEqual(amp.waveform_id.station_code, 'MDJ')
        self.assertEqual(amp.creation_info, None)

    def test_arrival(self):
        """
        Tests Arrival object.
        """
        self.assertEqual(len(self.catalog[0].origins[0].arrivals), 52)
        ar = self.catalog[0].origins[0].arrivals[0]
        self.assertEqual(ar.phase, 'Pn')
        self.assertEqual(ar.azimuth, 41.4)
        self.assertEqual(ar.distance, 2.22)
        self.assertEqual(ar.takeoff_angle, None)
        self.assertEqual(ar.takeoff_angle_errors.uncertainty, None)
        self.assertEqual(ar.time_residual, -1.9)
        self.assertEqual(ar.horizontal_slowness_residual, None)
        self.assertEqual(ar.backazimuth_residual, None)
        self.assertEqual(ar.time_weight, None)
        self.assertEqual(ar.horizontal_slowness_weight, None)
        self.assertEqual(ar.backazimuth_weight, None)
        self.assertEqual(
            ar.earth_model_id,
            ResourceIdentifier('quakeml:us.anss.org/earthmodel/ak135'))
        self.assertEqual(len(ar.comments), 0)

    def test_pick(self):
        """
        Tests Pick object.
        """
        self.assertEqual(len(self.catalog[0].picks), 52)
        pick = self.catalog[0].picks[0]
        self.assertEqual(pick.time, UTCDateTime(2012, 1, 1, 5, 28, 48, 180000))
        self.assertEqual(pick.time_errors.uncertainty, None)
        self.assertEqual(pick.waveform_id.station_code, 'JHJ2')
        self.assertAlmostEqual(pick.backazimuth, -138.6)
        self.assertEqual(pick.onset, 'emergent')
        self.assertEqual(pick.phase_hint, 'Pn')
        self.assertEqual(pick.polarity, None)
        self.assertEqual(pick.evaluation_mode, None)
        self.assertEqual(pick.evaluation_status, None)
        self.assertEqual(len(pick.comments), 0)

    def test_focalmechanism(self):
        """
        Tests FocalMechanism object.
        """
        self.assertEqual(len(self.catalog[0].focal_mechanisms), 4)
        fm = self.catalog[0].focal_mechanisms[0]
        self.assertEqual(
            fm.resource_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/focalmechanism/'
                   '20120101052755.98/ucmt/mwc'))
        # general
        self.assertEqual(fm.waveform_id, [])
        self.assertEqual(fm.triggering_origin_id, None)
        self.assertEqual(fm.azimuthal_gap, None)
        self.assertEqual(fm.station_polarity_count, None)
        self.assertEqual(fm.misfit, None)
        self.assertEqual(fm.station_distribution_ratio, None)
        self.assertEqual(
            fm.method_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/methodID=CMT'))
        # comments
        self.assertEqual(len(fm.comments), 0)
        # creation info
        self.assertEqual(fm.creation_info.author, None)
        self.assertEqual(fm.creation_info.agency_id, 'UCMT')
        self.assertEqual(fm.creation_info.author_uri, None)
        self.assertEqual(fm.creation_info.agency_uri, None)
        self.assertEqual(fm.creation_info.creation_time, None)
        self.assertEqual(fm.creation_info.version, None)
        # nodalPlanes
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_1.strike, 5.0)
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_1.dip, 85.0)
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_1.rake, -76.0)
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_2.strike, 116.0)
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_2.dip, 15.0)
        self.assertAlmostEqual(fm.nodal_planes.nodal_plane_2.rake, -159.0)
        self.assertEqual(fm.nodal_planes.preferred_plane, None)
        # principalAxes
        self.assertAlmostEqual(fm.principal_axes.t_axis.azimuth, 82.0)
        self.assertAlmostEqual(fm.principal_axes.t_axis.plunge, 38.0)
        self.assertAlmostEqual(fm.principal_axes.t_axis.length, 1.87e+19)
        self.assertAlmostEqual(fm.principal_axes.p_axis.azimuth, 290.0)
        self.assertAlmostEqual(fm.principal_axes.p_axis.plunge, 49.0)
        self.assertAlmostEqual(fm.principal_axes.p_axis.length, -1.87e+19)
        self.assertEqual(fm.principal_axes.n_axis.azimuth, 184)
        self.assertEqual(fm.principal_axes.n_axis.plunge, 14)
        self.assertEqual(fm.principal_axes.n_axis.length, 0.0)
        # momentTensor
        mt = fm.moment_tensor
        self.assertEqual(
            mt.resource_id,
            ResourceIdentifier(
                id='quakeml:us.anss.org/momenttensor/'
                   '20120101052755.98/ucmt/mwc'))
        self.assertAlmostEqual(mt.scalar_moment, 1.9e+19)
        self.assertAlmostEqual(mt.tensor.m_rr, -3.4e+18)
        self.assertAlmostEqual(mt.tensor.m_tt, -8e+17)
        self.assertAlmostEqual(mt.tensor.m_pp, 4.2e+18)
        self.assertAlmostEqual(mt.tensor.m_rt, -1.9e+18)
        self.assertAlmostEqual(mt.tensor.m_rp, -1.77e+19)
        self.assertAlmostEqual(mt.tensor.m_tp, -4.2e+18)
        self.assertEqual(mt.clvd, None)

    def test_write_quakeml(self):
        """
        Tests writing a QuakeML document.
        """
        with NamedTemporaryFile() as tf:
            _write_quakeml(self.catalog, tf, validate=IS_RECENT_LXML)
            # Read file again. Avoid the (legit) warning about the already used
            # resource identifiers.
            tf.seek(0)
            with warnings.catch_warnings(record=True):
                warnings.simplefilter("ignore")
                catalog2 = _read_quakeml(tf)
        self.assertTrue(len(catalog2), 1)

    def test_read_events(self):
        """
        Tests reading an mchedr document via read_events.
        """
        filename = os.path.join(self.path, 'mchedr.dat')
        # Read file again. Avoid the (legit) warning about the already used
        # resource identifiers.
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('always')
            catalog = read_events(filename)
            self.assertTrue(len(catalog), 1)


def suite():
    return unittest.makeSuite(MchedrTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
Python interface to the Seismic Analysis Code (SAC) file format.

:copyright:
    The Los Alamos National Security, LLC, Yannik Behr, C. J. Ammon,
    C. Satriano, L. Krischer, and J. MacCarthy
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


The SACTrace object maintains consistency between SAC headers and manages
header values in a user-friendly way. This includes some value-checking, native
Python logicals (True, False) and nulls (None) instead of SAC's 0, 1, or
-12345...

SAC headers are implemented as properties, with appropriate getters and
setters.


Features
--------

1. **Read and write SAC binary or ASCII**

   -  autodetect or specify expected byteorder
   -  optional file size checking and/or header consistency checks
   -  header-only reading and writing
   -  "overwrite OK" checking ('lovrok' header)

2. **Convenient access and manipulation of relative and absolute time
   headers**
3. **User-friendly header printing/viewing**
4. **Fast access to header values from attributes**

   -  With type checking, null handling, and enumerated value checking

5. **Convert to/from ObsPy Traces**

   -  Conversion from ObsPy Trace to SAC trace retains detected previous
      SAC header values.
   -  Conversion to ObsPy Trace retains the *complete* SAC header.


Usage examples
--------------

Read/write SAC files
~~~~~~~~~~~~~~~~~~~~

.. code:: python

    # read from a binary file
    sac = SACTrace.read(filename)

    # read header only
    sac = SACTrace.read(filename, headonly=True)

    # write header-only, file must exist
    sac.write(filename, headonly=True)

    # read from an ASCII file
    sac = SACTrace.read(filename, ascii=True)

    # write a binary SAC file for a Sun machine
    sac.write(filename, byteorder='big')

Build a SACTrace from a header dictionary and data array
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. rubric:: Example

>>> header = {'kstnm': 'ANMO', 'kcmpnm': 'BHZ', 'stla': 40.5, 'stlo': -108.23,
...           'evla': -15.123, 'evlo': 123, 'evdp': 50, 'nzyear': 2012,
...           'nzjday': 123, 'nzhour': 13, 'nzmin': 43, 'nzsec': 17,
...           'nzmsec': 100, 'delta': 1.0/40}
>>> sac = SACTrace(data=np.random.random(100), **header)
>>> print(sac)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
Reference Time = 05/02/2012 (123) 13:43:17.100000
   iztype IB: begin time
b          = 0.0
delta      = 0.0250000...
e          = 2.4750000...
evdp       = 50.0
evla       = -15.123000...
evlo       = 123.0
iftype     = itime
internal0  = 2.0
iztype     = ib
kcmpnm     = BHZ
kstnm      = ANMO
lcalda     = False
leven      = True
lovrok     = True
lpspol     = True
npts       = 100
nvhdr      = 6
nzhour     = 13
nzjday     = 123
nzmin      = 43
nzmsec     = 100
nzsec      = 17
nzyear     = 2012
stla       = 40.5
stlo       = -108.23000...

Reference-time and relative time headers
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. rubric:: Example

>>> sac = SACTrace(nzyear=2000, nzjday=1, nzhour=0, nzmin=0, nzsec=0,
...                nzmsec=0, t1=23.5, data=np.arange(100))
>>> print(sac.reftime)
2000-01-01T00:00:00.000000Z
>>> sac.b, sac.e, sac.t1
(0.0, 99.0, 23.5)

Move reference time by relative seconds, relative time headers are
preserved.

.. rubric:: Example

>>> sac = SACTrace(nzyear=2000, nzjday=1, nzhour=0, nzmin=0, nzsec=0,
...                nzmsec=0, t1=23.5, data=np.arange(100))
>>> sac.reftime -= 2.5
>>> sac.b, sac.e, sac.t1
(2.5, 101.5, 26.0)

Set reference time to new absolute time, relative time headers are
preserved.

.. rubric:: Example

>>> sac = SACTrace(nzyear=2000, nzjday=1, nzhour=0, nzmin=0, nzsec=0,
...                nzmsec=0, t1=23.5, data=np.arange(100))
>>> # set the reftime two minutes later
>>> sac.reftime = UTCDateTime(2000, 1, 1, 0, 2, 0, 0)
>>> sac.b, sac.e, sac.t1
(-120.0, -21.0, -96.5)

Quick header viewing
~~~~~~~~~~~~~~~~~~~~

Print non-null header values.

.. rubric:: Example

>>> sac = SACTrace()
>>> print(sac)  # doctest: +NORMALIZE_WHITESPACE
Reference Time = 01/01/1970 (001) 00:00:00.000000
   iztype IB: begin time
b          = 0.0
delta      = 1.0
e          = 0.0
iftype     = itime
internal0  = 2.0
iztype     = ib
lcalda     = False
leven      = True
lovrok     = True
lpspol     = True
npts       = 0
nvhdr      = 6
nzhour     = 0
nzjday     = 1
nzmin      = 0
nzmsec     = 0
nzsec      = 0
nzyear     = 1970

Print relative time header values.

.. rubric:: Example

>>> sac = SACTrace()
>>> sac.lh('picks')  # doctest: +NORMALIZE_WHITESPACE
Reference Time = 01/01/1970 (001) 00:00:00.000000
   iztype IB: begin time
a          = None
b          = 0.0
e          = 0.0
f          = None
o          = None
t0         = None
t1         = None
t2         = None
t3         = None
t4         = None
t5         = None
t6         = None
t7         = None
t8         = None
t9         = None

Header values as attributes
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Great for interactive use, with (ipython) tab-completion...

.. code:: python

    sac.<tab>

::

    sac.a                 sac.kevnm             sac.nzsec
    sac.az                sac.kf                sac.nzyear
    sac.b                 sac.khole             sac.o
    sac.baz               sac.kinst             sac.odelta
    sac.byteorder         sac.knetwk            sac.read
    sac.cmpaz             sac.ko                sac.reftime
    sac.cmpinc            sac.kstnm             sac.scale
    sac.copy              sac.kt0               sac.stdp
    sac.data              sac.kt1               sac.stel
    sac.delta             sac.kt2               sac.stla
    sac.depmax            sac.kt3               sac.stlo
    sac.depmen            sac.kt4               sac.t0
    sac.depmin            sac.kt5               sac.t1
    sac.dist              sac.kt6               sac.t2
    sac.e                 sac.kt7               sac.t3
    sac.evdp              sac.kt8               sac.t4
    sac.evla              sac.kt9               sac.t5
    sac.evlo              sac.kuser0            sac.t6
    sac.f                 sac.kuser1            sac.t7
    sac.from_obspy_trace  sac.kuser2            sac.t8
    sac.gcarc             sac.lcalda            sac.t9
    sac.idep              sac.leven             sac.to_obspy_trace
    sac.ievreg            sac.lh                sac.unused23
    sac.ievtyp            sac.listhdr           sac.user0
    sac.iftype            sac.lovrok            sac.user1
    sac.iinst             sac.lpspol            sac.user2
    sac.imagsrc           sac.mag               sac.user3
    sac.imagtyp           sac.nevid             sac.user4
    sac.internal0         sac.norid             sac.user5
    sac.iqual             sac.npts              sac.user6
    sac.istreg            sac.nvhdr             sac.user7
    sac.isynth            sac.nwfid             sac.user8
    sac.iztype            sac.nzhour            sac.user9
    sac.ka                sac.nzjday            sac.validate
    sac.kcmpnm            sac.nzmin             sac.write
    sac.kdatrd            sac.nzmsec

...and documentation (in IPython)!

.. code:: python

    sac.iztype?

::

    Type:        property
    String form: <property object at 0x106404940>
    Docstring:
    I    Reference time equivalence:

    * IUNKN (5): Unknown
    * IB (9): Begin time
    * IDAY (10): Midnight of reference GMT day
    * IO (11): Event origin time
    * IA (12): First arrival time
    * ITn (13-22): User defined time pick n, n=0,9

Convert to/from ObsPy Traces
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. rubric:: Example

>>> from obspy import read
>>> tr = read()[0]
>>> print(tr.stats)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
         network: BW
         station: RJOB
        location:
         channel: EHZ
       starttime: 2009-08-24T00:20:03.000000Z
         endtime: 2009-08-24T00:20:32.990000Z
   sampling_rate: 100.0
           delta: 0.01
            npts: 3000
           calib: 1.0
    back_azimuth: 100.0
     inclination: 30.0
        response: Channel Response
            ...


>>> sac = SACTrace.from_obspy_trace(tr)
>>> print(sac)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
Reference Time = 08/24/2009 (236) 00:20:03.000000
   iztype IB: begin time
b          = 0.0
delta      = 0.009999999...
e          = 29.989999...
iftype     = itime
iztype     = ib
kcmpnm     = EHZ
knetwk     = BW
kstnm      = RJOB
lcalda     = False
leven      = True
lovrok     = True
lpspol     = True
npts       = 3000
nvhdr      = 6
nzhour     = 0
nzjday     = 236
nzmin      = 20
nzmsec     = 0
nzsec      = 3
nzyear     = 2009
scale      = 1.0

>>> tr2 = sac.to_obspy_trace()
>>> print(tr2.stats)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
         network: BW
         station: RJOB
        location:
         channel: EHZ
       starttime: 2009-08-24T00:20:03.000000Z
         endtime: 2009-08-24T00:20:32.990000Z
   sampling_rate: 100.0
           delta: 0.01
            npts: 3000
           calib: 1.0
             sac: AttribDict(...)

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import sys
import warnings
from copy import deepcopy
from itertools import chain

import numpy as np
from obspy import Trace, UTCDateTime
from obspy.geodetics import gps2dist_azimuth, kilometer2degrees

from . import header as HD  # noqa
from .util import SacError, SacHeaderError
from . import util as _ut
from . import arrayio as _io


# ------------- HEADER DESCRIPTORS --------------------------------------------
#
# A descriptor is a class that manages an object attribute, using the
# descriptor protocol.  A single instance of a descriptor class (FloatHeader,
# for example) will exist for both the host class and all instances of that
# host class (i.e. SACTrace and all its instances).  As a result, we must
# implement logic that can tell if the methods are being called on the host
# class or on an instance.  This looks like "if instance is None" on methods.
#
# See:
# https://docs.python.org/3.5/howto/descriptor.html
# https://nbviewer.jupyter.org/urls/gist.github.com/ChrisBeaumont/
#   5758381/raw/descriptor_writeup.ipynb

class SACHeader(object):
    def __init__(self, name):
        try:
            self.__doc__ = HD.DOC[name]
        except KeyError:
            # header doesn't have a docstring entry in HD.DOC
            pass
        self.name = name


class FloatHeader(SACHeader):
    def __get__(self, instance, instance_type):
        if instance is None:
            # a FloatHeader on the owner class was requested.
            # return the descriptor itself.
            value = self
        else:
            # a FloatHeader on an instance was requested.
            # return the descriptor value.
            value = float(instance._hf[HD.FLOATHDRS.index(self.name)])
            if value == HD.FNULL:
                value = None
        return value

    def __set__(self, instance, value):
        if value is None:
            value = HD.FNULL
        instance._hf[HD.FLOATHDRS.index(self.name)] = value


# Descriptor for setting relative time headers with either a relative
# time float or an absolute UTCDateTime
# used for: b, o, a, f, t0-t9
class RelativeTimeHeader(FloatHeader):
    def __set__(self, instance, value):
        """
        Intercept the set value to make sure it is an offset from the SAC
        reference time.

        """
        if isinstance(value, UTCDateTime):
            offset = value - instance.reftime
        else:
            offset = value
        # reuse the normal floatheader setter.
        super(RelativeTimeHeader, self).__set__(instance, offset)


# Factory function for setting geographic header values
#   (evlo, evla, stalo, stalat)
# that will check lcalda and calculate and set dist, az, baz, gcarc
class GeographicHeader(FloatHeader):
    def __set__(self, instance, value):
        super(GeographicHeader, self).__set__(instance, value)
        if instance.lcalda:
            try:
                instance._set_distances()
            except SacHeaderError:
                pass


class IntHeader(SACHeader):
    def __get__(self, instance, instance_type):
        if instance is None:
            value = self
        else:
            value = int(instance._hi[HD.INTHDRS.index(self.name)])
            if value == HD.INULL:
                value = None
        return value

    def __set__(self, instance, value):
        if value is None:
            value = HD.INULL
        if value % 1:
            warnings.warn("Non-integers may be truncated. ({}: {})".format(
                self.name, value))
        instance._hi[HD.INTHDRS.index(self.name)] = value


class BoolHeader(IntHeader):
    def __get__(self, instance, instance_type):
        # value can be an int or None
        value = super(BoolHeader, self).__get__(instance, instance_type)
        return bool(value) if value in (0, 1) else value

    def __set__(self, instance, value):
        if value not in (True, False, 1, 0):
            msg = "Logical header values must be {True, False, 1, 0}"
            raise ValueError(msg)
        # booleans are subclasses of integers.  They will be set (cast)
        # directly into an integer array as 0 or 1.
        super(BoolHeader, self).__set__(instance, value)
        if self.name == 'lcalda':
            if value:
                try:
                    instance._set_distances()
                except SacHeaderError:
                    pass


class EnumHeader(IntHeader):
    def __get__(self, instance, instance_type):
        value = super(EnumHeader, self).__get__(instance, instance_type)
        # value is int or None
        if value is None:
            name = None
        elif _ut.is_valid_enum_int(self.name, value):
            name = HD.ENUM_NAMES[value]
        else:
            msg = """Unrecognized enumerated value {} for header "{}".
                     See .header for allowed values.""".format(value,
                                                               self.name)
            warnings.warn(msg)
            name = None
        return name

    def __set__(self, instance, value):
        if value is None:
            value = HD.INULL
        elif _ut.is_valid_enum_str(self.name, value):
            if self.name == 'iztype':
                reftime = _iztype_reftime(instance, value)
                instance.reftime = reftime
                # this also shifts all non-null relative times (instance._allt)
            value = HD.ENUM_VALS[value]
        else:
            msg = 'Unrecognized enumerated value "{}" for header "{}"'
            raise ValueError(msg.format(value, self.name))
        super(EnumHeader, self).__set__(instance, value)


class StringHeader(SACHeader):
    def __get__(self, instance, instance_type):
        if instance is None:
            value = self
        else:
            value = instance._hs[HD.STRHDRS.index(self.name)]
            try:
                # value is a bytes
                value = value.decode()
            except AttributeError:
                # value is a str
                pass

            if value == HD.SNULL:
                value = None

            try:
                value = value.strip()
            except AttributeError:
                # it's None.  no .strip method
                pass
        return value

    def __set__(self, instance, value):
        if value is None:
            value = HD.SNULL
        elif len(value) > 8:
            msg = ("Alphanumeric headers longer than 8 characters are "
                   "right-truncated.")
            warnings.warn(msg)
        # values will truncate themselves, since _hs is dtype '|S8'
        try:
            instance._hs[HD.STRHDRS.index(self.name)] = value.encode('ascii',
                                                                     'strict')
        except AttributeError:
            instance._hs[HD.STRHDRS.index(self.name)] = value


# Headers for functions of .data (min, max, mean, len)
class DataHeader(SACHeader):
    def __init__(self, name, func):
        self.func = func
        super(DataHeader, self).__init__(name)

    def __get__(self, instance, instance_type):
        if instance is None:
            value = self
        else:
            try:
                value = self.func(instance.data)
                # convert to native Python types
                if self.name in HD.FLOATHDRS:
                    value = float(value)
                elif self.name in HD.INTHDRS:
                    value = int(value)
            except TypeError:
                # instance.data is None, get value from header
                if self.name in HD.FLOATHDRS:
                    value = instance._hf[HD.FLOATHDRS.index(self.name)].item()
                    value = None if value == HD.FNULL else value
                elif self.name in HD.INTHDRS:
                    value = instance._hi[HD.INTHDRS.index(self.name)].item()
                    value = None if value == HD.INULL else value
        return value

    def __set__(self, instance, value):
        msg = "{} is read-only".format(self.name)
        raise AttributeError(msg)


# OTHER GETTERS/SETTERS
def _get_e(self):
    try:
        if self.npts:
            e = self.b + (self.npts - 1) * self.delta
        else:
            e = self.b
    except TypeError:
        # b, npts, and/or delta are None/null
        # TODO: assume "b" is 0.0?
        e = None
    return e


def _iztype_reftime(sactrace, iztype):
    """
    Get the new reftime for a given iztype.

    Setting the iztype will shift the relative time headers, such that the
    header that iztype points to is (near) zero, and all others are shifted
    together by the difference.

    Affected headers: b, o, a, f, t0-t9

    :param sactrace:
    :type sactrace: SACTrace
    :param iztype: One of the following strings:
        'iunkn'
        'ib', begin time
        'iday', midnight of reference GMT day
        'io', event origin time
        'ia', first arrival time
        'it0'-'it9', user defined pick t0-t9.
    :type iztype: str
    :rtype reftime: UTCDateTime
    :return: The new SAC reference time.

    """
    # The Plan:
    # 1. find the seconds needed to shift the old reftime to the new one.
    # 2. shift reference time onto the iztype header using that shift value.
    # 3. this triggers an _allt shift of all relative times by the same amount.
    # 4. If all goes well, actually set the iztype in the header.

    # 1.
    if iztype == 'iunkn':
        # no shift
        ref_val = 0.0
    elif iztype == 'iday':
        # seconds since midnight of reference day
        reftime = sactrace.reftime
        ref_val = reftime - UTCDateTime(year=reftime.year,
                                        julday=reftime.julday)
    else:
        # a relative time header.
        # remove the 'i' (first character) in the iztype to get the header name
        ref_val = getattr(sactrace, iztype[1:])
        if ref_val is None:
            msg = "Reference header for iztype '{}' is not set".format(iztype)
            raise SacError(msg)

    # 2. set a new reference time,
    # 3. which also shifts all non-null relative times (sactrace._allt).
    #    remainder microseconds may be in the reference header value, because
    #    nzmsec can't hold them.
    new_reftime = sactrace.reftime + ref_val

    return new_reftime


# kevnm is 16 characters, split into two 8-character fields
# intercept and handle in while getting and setting
def _get_kevnm(self):
    kevnm = self._hs[HD.STRHDRS.index('kevnm')]
    kevnm2 = self._hs[HD.STRHDRS.index('kevnm2')]
    try:
        kevnm = kevnm.decode()
        kevnm2 = kevnm2.decode()
    except AttributeError:
        # kevnm is a str
        pass

    if kevnm == HD.SNULL:
        kevnm = ''
    if kevnm2 == HD.SNULL:
        kevnm2 = ''

    value = (kevnm + kevnm2).strip()

    if not value:
        value = None

    return value


def _set_kevnm(self, value):
    if value is None:
        value = HD.SNULL + HD.SNULL
    elif len(value) > 16:
        msg = "kevnm over 16 characters.  Truncated to {}.".format(value[:16])
        warnings.warn(msg)
    kevnm = '{:<8s}'.format(value[0:8])
    kevnm2 = '{:<8s}'.format(value[8:16])
    self._hs[HD.STRHDRS.index('kevnm')] = kevnm
    self._hs[HD.STRHDRS.index('kevnm2')] = kevnm2

# TODO: move get/set reftime up here, make it a property


# -------------------------- SAC OBJECT INTERFACE -----------------------------
class SACTrace(object):
    __doc__ = """
    Convenient and consistent in-memory representation of Seismic Analysis Code
    (SAC) files.

    This is the human-facing interface for making a valid instance.  For
    file-based or other constructors, see class methods .read and
    .from_obspy_trace.  SACTrace instances preserve relationships between
    header values.

    :param data: Associated time-series data vector. Optional. If omitted, None
        is set as the instance data attribute.
    :type data: :class:`numpy.ndarray` of float32

    Any valid header key/value pair is also an optional input keyword argument.
    If not provided, minimum required headers are set to valid default values.
    The default instance is an evenly-space trace, with a sample rate of 1.0,
    and len(data) or 0 npts, starting at 1970-01-01T00:00:00.000000.

    :var reftime: Read-only reference time.  Calculated from nzyear, nzjday,
        nzhour, nzmin, nzsec, nzmsec.
    :var byteorder: The byte order of the underlying header/data arrays.
        Raises :class:`SacError` if array byte orders are inconsistent, even in
        the case where '<' is your native order and byteorders look like '<',
        '=', '='.

    Any valid header name is also an attribute. See below, :mod:`header`,
    or individial attribution docstrings for more header information.

                                 THE SAC HEADER

    NOTE: All header names and string values are lowercase. Header value
    access should be through instance attributes.

    """ + HD.HEADER_DOCSTRING

    # ------------------------------- SAC HEADERS -----------------------------
    # SAC header values are defined as managed attributes, either as
    # descriptors or as properties, with getters and setters.
    #
    # Managed attributes are defined at the class leval, and therefore shared
    # across all instances, not attribute data themselves.
    #
    # This section looks ugly, but it allows for the following:
    # 1. Relationships/checks between header variables can be done in setters
    # 2. The underlying header array structure is retained, for quick writing.
    # 3. Header access looks like simple attribute-access syntax.
    #    Looks funny to read here, but natural to use.
    #
    # FLOATS
    delta = FloatHeader('delta')
    depmin = DataHeader('depmin', min)
    depmax = DataHeader('depmax', max)
    scale = FloatHeader('scale')
    odelta = FloatHeader('odelta')
    b = RelativeTimeHeader('b')
    e = property(_get_e, doc=HD.DOC['e'])
    o = RelativeTimeHeader('o')
    a = RelativeTimeHeader('a')
    internal0 = FloatHeader('internal0')
    t0 = RelativeTimeHeader('t0')
    t1 = RelativeTimeHeader('t1')
    t2 = RelativeTimeHeader('t2')
    t3 = RelativeTimeHeader('t3')
    t4 = RelativeTimeHeader('t4')
    t5 = RelativeTimeHeader('t5')
    t6 = RelativeTimeHeader('t6')
    t7 = RelativeTimeHeader('t7')
    t8 = RelativeTimeHeader('t8')
    t9 = RelativeTimeHeader('t9')
    f = RelativeTimeHeader('f')
    stla = GeographicHeader('stla')
    stlo = GeographicHeader('stlo')
    stel = FloatHeader('stel')
    stdp = FloatHeader('stdp')
    evla = GeographicHeader('evla')
    evlo = GeographicHeader('evlo')
    evdp = FloatHeader('evdp')
    mag = FloatHeader('mag')
    user0 = FloatHeader('user0')
    user1 = FloatHeader('user1')
    user2 = FloatHeader('user2')
    user3 = FloatHeader('user3')
    user4 = FloatHeader('user4')
    user5 = FloatHeader('user5')
    user6 = FloatHeader('user6')
    user7 = FloatHeader('user7')
    user8 = FloatHeader('user8')
    user9 = FloatHeader('user9')
    dist = FloatHeader('dist')
    az = FloatHeader('az')
    baz = FloatHeader('baz')
    gcarc = FloatHeader('gcarc')
    depmen = DataHeader('depmen', np.mean)
    cmpaz = FloatHeader('cmpaz')
    cmpinc = FloatHeader('cmpinc')
    #
    # INTS
    nzyear = IntHeader('nzyear')
    nzjday = IntHeader('nzjday')
    nzhour = IntHeader('nzhour')
    nzmin = IntHeader('nzmin')
    nzsec = IntHeader('nzsec')
    nzmsec = IntHeader('nzmsec')
    nvhdr = IntHeader('nvhdr')
    norid = IntHeader('norid')
    nevid = IntHeader('nevid')
    npts = DataHeader('npts', len)
    nwfid = IntHeader('nwfid')
    iftype = EnumHeader('iftype')
    idep = EnumHeader('idep')
    iztype = EnumHeader('iztype')
    iinst = IntHeader('iinst')
    istreg = IntHeader('istreg')
    ievreg = IntHeader('ievreg')
    ievtyp = EnumHeader('ievtyp')
    iqual = IntHeader('iqual')
    isynth = EnumHeader('isynth')
    imagtyp = EnumHeader('imagtyp')
    imagsrc = EnumHeader('imagsrc')
    leven = BoolHeader('leven')
    lpspol = BoolHeader('lpspol')
    lovrok = BoolHeader('lovrok')
    lcalda = BoolHeader('lcalda')
    unused23 = IntHeader('unused23')
    #
    # STRINGS
    kstnm = StringHeader('kstnm')
    kevnm = property(_get_kevnm, _set_kevnm, doc=HD.DOC['kevnm'])
    khole = StringHeader('khole')
    ko = StringHeader('ko')
    ka = StringHeader('ka')
    kt0 = StringHeader('kt0')
    kt1 = StringHeader('kt1')
    kt2 = StringHeader('kt2')
    kt3 = StringHeader('kt3')
    kt4 = StringHeader('kt4')
    kt5 = StringHeader('kt5')
    kt6 = StringHeader('kt6')
    kt7 = StringHeader('kt7')
    kt8 = StringHeader('kt8')
    kt9 = StringHeader('kt9')
    kf = StringHeader('kf')
    kuser0 = StringHeader('kuser0')
    kuser1 = StringHeader('kuser1')
    kuser2 = StringHeader('kuser2')
    kcmpnm = StringHeader('kcmpnm')
    knetwk = StringHeader('knetwk')
    kdatrd = StringHeader('kdatrd')
    kinst = StringHeader('kinst')

    def __init__(self, leven=True, delta=1.0, b=0.0, e=0.0, iztype='ib',
                 nvhdr=6, npts=0, iftype='itime', nzyear=1970, nzjday=1,
                 nzhour=0, nzmin=0, nzsec=0, nzmsec=0, lcalda=False,
                 lpspol=True, lovrok=True, internal0=2.0, data=None, **kwargs):
        """
        Initialize a SACTrace object using header key-value pairs and a
        numpy.ndarray for the data, both optional.

        ..rubric:: Example

        >>> sac = SACTrace(nzyear=1995, nzmsec=50, data=np.arange(100))
        >>> print(sac)  # doctest: +NORMALIZE_WHITESPACE
        Reference Time = 01/01/1995 (001) 00:00:00.050000
           iztype IB: begin time
        b          = 0.0
        delta      = 1.0
        e          = 99.0
        iftype     = itime
        internal0  = 2.0
        iztype     = ib
        lcalda     = False
        leven      = True
        lovrok     = True
        lpspol     = True
        npts       = 100
        nvhdr      = 6
        nzhour     = 0
        nzjday     = 1
        nzmin      = 0
        nzmsec     = 50
        nzsec      = 0
        nzyear     = 1995

        """
        # The Plan:
        # 1. Build the default header dictionary and update with provided
        #    values.
        # 2. Convert header dict to arrays (util.dict_to_header_arrays
        #    initializes the arrays and fills in without checking.
        # 3. set the _h[fis] and data arrays on self.

        # 1.
        # build the required header from provided or default values
        header = {'leven': leven, 'npts': npts, 'delta': delta, 'b': b, 'e': e,
                  'iztype': iztype, 'nvhdr': nvhdr, 'iftype': iftype,
                  'nzyear': nzyear, 'nzjday': nzjday, 'nzhour': nzhour,
                  'nzmin': nzmin, 'nzsec': nzsec, 'nzmsec': nzmsec,
                  'lcalda': lcalda, 'lpspol': lpspol, 'lovrok': lovrok,
                  'internal0': internal0}

        # combine header with remaining non-required args.
        # user can put non-SAC key:value pairs into the header, but they're
        # ignored on write.
        header.update(kwargs)

        # -------------------------- DATA ARRAY -------------------------------
        if data is None:
            # this is like "headonly=True"
            pass
        else:
            if not isinstance(data, np.ndarray):
                raise TypeError("data needs to be a numpy.ndarray")
            else:
                # Only copy the data if they are not of the required type
                # XXX: why require little endian instead of native byte order?
                # data = np.require(data, native_str('<f4'))
                pass

        # --------------------------- HEADER ARRAYS ---------------------------
        # 2.
        # TODO: this is done even when we're reading a file.
        #   if it's too much overhead, it may need to change

        # swap enum names for integer values in the header dictionary
        header = _ut.enum_string_to_int(header)

        # XXX: will these always be little endian?
        hf, hi, hs = _io.dict_to_header_arrays(header)

        # we now have data and headers, either default or provided.

        # 3.
        # this completely sidesteps any checks provided by class properties
        self._hf = hf
        self._hi = hi
        self._hs = hs
        self.data = data

        self._set_distances()

    @property
    def _header(self):
        """
        Convenient read-only dictionary of non-null header array values.

        Header value access should be through instance attributes. All header
        names and string values are lowercase. Computed every time, so use
        frugally. See class docstring for header descriptions.

        """
        out = _io.header_arrays_to_dict(self._hf, self._hi, self._hs,
                                        nulls=False)
        return out

    @property
    def byteorder(self):
        """
        The byte order of the underlying header/data arrays.

        Raises SacError if array byte orders are inconsistent, even in the
        case where '<' is your native order and byteorders look like
        '<', '=', '='.

        """
        try:
            if self.data is None:
                assert self._hf.dtype.byteorder == self._hi.dtype.byteorder
            else:
                assert self._hf.dtype.byteorder == self._hi.dtype.byteorder ==\
                    self.data.dtype.byteorder
        except AssertionError:
            msg = 'Inconsistent header/data byteorders.'
            raise SacError(msg)

        bo = self._hf.dtype.byteorder

        if bo == '=':
            byteorder = sys.byteorder
        elif bo == '<':
            byteorder = 'little'
        elif bo == '>':
            byteorder = 'big'

        return byteorder

    # TODO: make a byteorder setter?
    def _byteswap(self):
        """
        Change the underlying byte order and dtype interpretation of the float,
        int, and (if present) data arrays.

        """
        try:
            self._hf = self._hf.byteswap(True).newbyteorder('S')
            self._hi = self._hi.byteswap(True).newbyteorder('S')
            if self.data is not None:
                self.data = self.data.byteswap(True).newbyteorder('S')
        except Exception as e:
            # if this fails, roll it back?
            raise e

    @property
    def reftime(self):
        """
        Get or set the SAC header reference time as a UTCDateTime instance.

        reftime is not an attribute, but is constructed and dismantled each
        time directly to/from the SAC "nz"-time fields.

        Setting a new reftime shifts all non-null relative time headers
        accordingly.  It accepts a UTCDateTime object, from which time shifts
        are calculated.

        ..rubric:: notes

        The reftime you supply will be robbed of its remainder microseconds,
        which are then pushed into the relative time header shifts.  This means
        that the reftime you observe after you set it here may not exactly
        match the reftime you supplied; it may be `remainder microseconds`
        earlier. Nor will the iztype reference header value be exactly zero;
        it will be equal to `remainder microseconds` (as seconds).

        """
        return _ut.get_sac_reftime(self._header)

    @reftime.setter
    def reftime(self, new_reftime):
        try:
            old_reftime = self.reftime

            # snap the new reftime to the most recent milliseconds
            # (subtract the leftover microseconds)
            ns = new_reftime.ns
            utc = UTCDateTime(ns=(ns - ns % 1000000))

            self.nzyear = utc.year
            self.nzjday = utc.julday
            self.nzhour = utc.hour
            self.nzmin = utc.minute
            self.nzsec = utc.second
            self.nzmsec = utc.microsecond / 1000

            # get the float seconds between the old and new reftimes
            shift = old_reftime - utc

            # shift the relative time headers
            self._allt(np.float32(shift))

        except AttributeError:
            msg = "New reference time must be an obspy.UTCDateTime instance."
            raise TypeError(msg)

    # --------------------------- I/O METHODS ---------------------------------
    @classmethod
    def read(cls, source, headonly=False, ascii=False, byteorder=None,
             checksize=False, debug_strings=False, encoding='ASCII'):
        """
        Construct an instance from a binary or ASCII file on disk.

        :param source: Full path string for File-like object from a SAC binary
            file on disk.  If it is an open File object, open 'rb'.
        :type source: str or file
        :param headonly: If headonly is True, only read the header arrays not
            the data array.
        :type headonly: bool
        :param ascii: If True, file is a SAC ASCII/Alphanumeric file.
        :type ascii: bool
        :param byteorder: If omitted or None, automatic byte-order checking is
            done, starting with native order. If byteorder is specified and
            incorrect, a :class:`SacIOError` is raised. Only valid for binary
            files.
        :type byteorder: str {'little', 'big'}, optional
        :param checksize: If True, check that the theoretical file size from
            the header matches the size on disk. Only valid for binary files.
        :type checksize: bool
        :param debug_strings: By default, non-ASCII and null-termination
            characters are removed from character header fields, and those
            beginning with '-12345' are considered unset. If True, they
            are instead passed without modification.  Good for debugging.
        :type debug_strings: bool
        :param encoding: Encoding string that passes the user specified
        encoding scheme.
        :type encoding: str

        :raises: :class:`SacIOError` if checksize failed, byteorder was wrong,
            or header arrays are wrong size.

        .. rubric:: Example

        >>> from obspy.core.util import get_example_file
        >>> from obspy.io.sac.util import SacInvalidContentError
        >>> file_ = get_example_file("test.sac")
        >>> sac = SACTrace.read(file_, headonly=True)
        >>> sac.data is None
        True
        >>> sac = SACTrace.read(file_, headonly=False)
        >>> sac.data  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        array([ -8.74227766e-08,  -3.09016973e-01,  -5.87785363e-01,
                -8.09017122e-01,  -9.51056600e-01,  -1.00000000e+00,
                -9.51056302e-01,  -8.09016585e-01,  -5.87784529e-01,
                ...
                 8.09022486e-01,   9.51059461e-01,   1.00000000e+00,
                 9.51053500e-01,   8.09011161e-01,   5.87777138e-01,
                 3.09007347e-01], dtype=float32)

        See also: :meth:`SACTrace.validate`

        """
        if ascii:
            hf, hi, hs, data = _io.read_sac_ascii(source, headonly=headonly)
        else:
            hf, hi, hs, data = _io.read_sac(source, headonly=headonly,
                                            byteorder=byteorder,
                                            checksize=checksize)
        if not debug_strings:
            for i, val in enumerate(hs):
                val = _ut._clean_str(val.decode(encoding, 'replace'),
                                     strip_whitespace=False)
                if val.startswith(native_str('-12345')):
                    val = HD.SNULL
                hs[i] = val.encode(encoding, 'replace')

        sac = cls._from_arrays(hf, hi, hs, data)
        if sac.dist is None:
            sac._set_distances()

        return sac

    def write(self, dest, headonly=False, ascii=False, byteorder=None,
              flush_headers=True):
        """
        Write the header and (optionally) data arrays to a SAC binary file.

        :param dest: Full path or File-like object to SAC binary file on disk.
        :type dest: str or file
        :param headonly: If headonly is True, only read the header arrays not
            the data array.
        :type headonly: bool
        :param ascii: If True, file is a SAC ASCII/Alphanumeric file.
        :type ascii: bool
        :param byteorder: If omitted or None, automatic byte-order checking is
            done, starting with native order. If byteorder is specified and
            incorrect, a :class:`SacIOError` is raised. Only valid for binary
            files.
        :type byteorder: str {'little', 'big'}, optional
        :param flush_headers: If True, update data headers like 'depmin' and
            'depmax' with values from the data array.
        :type flush_headers: bool

        """
        if headonly:
            data = None
        else:
            # do a check for float32 data here instead of arrayio.write_sac?
            data = self.data
            if flush_headers:
                self._flush_headers()

        if ascii:
            _io.write_sac_ascii(dest, self._hf, self._hi, self._hs, data)
        else:
            byteorder = byteorder or self.byteorder
            _io.write_sac(dest, self._hf, self._hi, self._hs, data,
                          byteorder=byteorder)

    @classmethod
    def _from_arrays(cls, hf=None, hi=None, hs=None, data=None):
        """
        Low-level array-based constructor.

        This constructor is good for getting a "blank" SAC object, and is used
        in other, perhaps more useful, alternate constructors ("See Also").
        No value checking is done and header values are completely overwritten
        with the provided arrays, which is why this is a hidden constructor.

        :param hf: SAC float header array
        :type hf: :class:`numpy.ndarray` of floats
        :param hi: SAC int header array
        :type hi: :class:`numpy.ndarray` of ints
        :param hs: SAC string header array
        :type hs: :class:`numpy.ndarray` of str
        :param data: SAC data array, optional.

        If omitted or None, the header arrays are intialized according to
        :func:`arrayio.init_header_arrays`.  If data is omitted, it is
        simply set to None on the corresponding :class:`SACTrace`.

        .. rubric:: Example

        >>> sac = SACTrace._from_arrays()
        >>> print(sac)  # doctest: +NORMALIZE_WHITESPACE +SKIP
        Reference Time = XX/XX/XX (XXX) XX:XX:XX.XXXXXX
            iztype not set
        lcalda     = True
        leven      = False
        lovrok     = False
        lpspol     = False

        """
        # use the first byteorder we find, or system byteorder if we
        # never find any
        bo = '='
        for arr in (hf, hi, hs, data):
            try:
                bo = arr.dtype.byteorder
                break
            except AttributeError:
                # arr is None (not supplied)
                pass
        hf0, hi0, hs0 = _io.init_header_arrays(byteorder=bo)
        # TODO: hf0, hi0, hs0 = _io.init_header_array_values(hf0, hi0, hs0)

        if hf is None:
            hf = hf0
        if hi is None:
            hi = hi0
        if hs is None:
            hs = hs0

        # get the default instance, but completely replace the arrays
        # initializes arrays twice, but it beats converting empty arrays to a
        # dict and then passing it to __init__, i think...maybe...
        sac = cls()
        sac._hf = hf
        sac._hi = hi
        sac._hs = hs
        sac.data = data

        return sac

    # TO/FROM OBSPY TRACES
    @classmethod
    def from_obspy_trace(cls, trace, keep_sac_header=True):
        """
        Construct an instance from an ObsPy Trace.

        :param trace: Source Trace object
        :type trace: :class:`~obspy.core.Trace` instance
        :param keep_sac_header: If True, any old stats.sac header values are
            kept as is, and only a minimal set of values are updated from the
            stats dictionary: npts, e, and data.  If an old iztype and a valid
            reftime are present, b and e will be properly referenced to it.  If
            False, a new SAC header is constructed from only information found
            in the stats dictionary, with some other default values introduced.
        :type keep_sac_header: bool

        """
        header = _ut.obspy_to_sac_header(trace.stats, keep_sac_header)

        # handle the data headers
        data = trace.data
        try:
            if len(data) == 0:
                # data is a empty numpy.array
                data = None
        except TypeError:
            # data is None
            data = None

        try:
            byteorder = data.dtype.byteorder
        except AttributeError:
            # data is None
            byteorder = '='

        hf, hi, hs = _io.dict_to_header_arrays(header, byteorder=byteorder)
        sac = cls._from_arrays(hf, hi, hs, data)
        # sac._flush_headers()

        return sac

    def to_obspy_trace(self, debug_headers=False, encoding='ASCII'):
        """
        Return an ObsPy Trace instance.

        Required headers: nz-time fields, npts, delta, calib, kcmpnm, kstnm,
        ...?

        :param debug_headers: Include _all_ SAC headers into the
            Trace.stats.sac dictionary.
        :type debug_headers: bool
        :param encoding: Encoding string that passes the user specified
        encoding scheme.
        :type encoding: str

        .. rubric:: Example

        >>> from obspy.core.util import get_example_file
        >>> file_ = get_example_file("test.sac")
        >>> sac = SACTrace.read(file_, headonly=True)
        >>> tr = sac.to_obspy_trace()
        >>> print(tr)  # doctest: +ELLIPSIS
        .STA..Q | 1978-07-18T08:00:10.000000Z - ... | 1.0 Hz, 100 samples

        """
        # make the obspy test for tests/data/testxy.sac pass
        # ObsPy does not require a valid reftime
        # try:
        #     self.validate('reftime')
        # except SacInvalidContentError:
        #     if not self.nzyear:
        #         self.nzyear = 1970
        #     if not self.nzjday:
        #         self.nzjday = 1
        #     for hdr in ['nzhour', 'nzmin', 'nzsec', 'nzmsec']:
        #         if not getattr(self, hdr):
        #             setattr(self, hdr, 0)
        self.validate('delta')
        if self.data is None:
            # headonly is True
            # Make it something palatable to ObsPy
            data = np.array([], dtype=self._hf.dtype.byteorder + 'f4')
        else:
            data = self.data

        sachdr = _io.header_arrays_to_dict(self._hf, self._hi, self._hs,
                                           nulls=debug_headers,
                                           encoding=encoding)
        # TODO: logic to use debug_headers for real

        stats = _ut.sac_to_obspy_header(sachdr)

        return Trace(data=data, header=stats)

    # ---------------------- other properties/methods -------------------------
    def validate(self, *tests):
        """
        Check validity of loaded SAC file content, such as header/data
        consistency.

        :param tests: One or more of the following validity tests:
            'delta' : Time step "delta" is positive.
            'logicals' : Logical values are 0, 1, or null
            'data_hdrs' : Length, min, mean, max of data array match header
                values.
            'enums' : Check validity of enumerated values.
            'reftime' : Reference time values in header are all set.
            'reltime' : Relative time values in header are absolutely
                referenced.
            'all' : Do all tests.
        :type tests: str

        :raises: :class:`SacInvalidContentError` if any of the specified tests
            fail. :class:`ValueError` if 'data_hdrs' is specified and data is
            None, empty array, or no tests specified.

        .. rubric:: Example

        >>> from obspy.core.util import get_example_file
        >>> from obspy.io.sac.util import SacInvalidContentError
        >>> file_ = get_example_file("LMOW.BHE.SAC")
        >>> sac = SACTrace.read(file_)
        >>> # make the time step invalid, catch it, and fix it
        >>> sac.delta *= -1.0
        >>> try:
        ...     sac.validate('delta')
        ... except SacInvalidContentError as e:
        ...     sac.delta *= -1.0
        ...     sac.validate('delta')
        >>> # make the data and depmin/men/max not match, catch the validation
        >>> # error, then fix (flush) the headers so that they validate
        >>> sac.data += 5.0
        >>> try:
        ...     sac.validate('data_hdrs')
        ... except SacInvalidContentError:
        ...     sac._flush_headers()
        ...     sac.validate('data_hdrs')

        """
        _io.validate_sac_content(self._hf, self._hi, self._hs, self.data,
                                 *tests)

    def _format_header_str(self, hdrlist='all'):
        """
        Produce a print-friendly string of header values for __repr__ ,
        .listhdr(), and .lh()

        """
        # interpret hdrlist
        if hdrlist == 'all':
            hdrlist = sorted(self._header.keys())
        elif hdrlist == 'picks':
            hdrlist = ('a', 'b', 'e', 'f', 'o', 't0', 't1', 't2', 't3', 't4',
                       't5', 't6', 't7', 't8', 't9')
        else:
            msg = "Unrecognized hdrlist '{}'".format(hdrlist)
            raise ValueError(msg)

        # start building header string
        #
        # reference time
        header_str = []
        try:
            timefmt = "Reference Time = %m/%d/%Y (%j) %H:%M:%S.%f"
            header_str.append(self.reftime.strftime(timefmt))
        except (ValueError, SacError):
            msg = "Reference time information incomplete."
            warnings.warn(msg)
            notime_str = "Reference Time = XX/XX/XX (XXX) XX:XX:XX.XXXXXX"
            header_str.append(notime_str)
        #
        # reftime type
        # TODO: use enumerated value dict here?
        iztype = self.iztype
        if iztype is None:
            header_str.append("\tiztype not set")
        elif iztype == 'ib':
            header_str.append("\tiztype IB: begin time")
        elif iztype == 'io':
            header_str.append("\tiztype IO: origin time")
        elif iztype == 'ia':
            header_str.append("\tiztype IA: first arrival time")
        elif iztype[1] == 't':
            vals = (iztype.upper(), iztype[1:])
            izfmt = "\tiztype {}: user-defined time {}"
            header_str.append(izfmt.format(*vals))
        elif iztype == 'iunkn':
            header_str.append("\tiztype IUNKN (Unknown)")
        else:
            header_str.append("\tunrecognized iztype: {}".format(iztype))
        #
        # non-null headers
        hdrfmt = "{:10.10s} = {}"
        for hdr in hdrlist:
            # XXX: non-null header values might have no property for getattr
            try:
                header_str.append(hdrfmt.format(hdr, getattr(self, hdr)))
            except AttributeError:
                header_str.append(hdrfmt.format(hdr, self._header[hdr]))

        return '\n'.join(header_str)

    def listhdr(self, hdrlist='all'):
        """
        Print header values.

        Default is all non-null values.

        :param hdrlist: Which header fields to you want to list. Choose one of
            {'all', 'picks'} or iterable of header fields.  An iterable of
            header fields can look like 'bea' or ('b', 'e', 'a').

            'all' (default) prints all non-null values.
            'picks' prints fields which are used to define time picks.

        An iterable of header fields can look like 'bea' or ('b', 'e', 'a').

        .. rubric:: Example

        >>> from obspy.core.util import get_example_file
        >>> file_ = get_example_file("LMOW.BHE.SAC")
        >>> sac = SACTrace.read(file_)
        >>> sac.lh()  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        Reference Time = 04/10/2001 (100) 00:23:00.465000
           iztype IB: begin time
        a          = 0.0
        b          = 0.0
        delta      = 0.009999999...
        depmax     = 0.003305610...
        depmen     = 0.00243799...
        depmin     = 0.00148824...
        e          = 0.98999997...
        iftype     = itime
        iztype     = ib
        kcmpnm     = BHE
        kevnm      = None
        kstnm      = LMOW
        lcalda     = True
        leven      = True
        lpspol     = False
        nevid      = 0
        norid      = 0
        npts       = 100
        nvhdr      = 6
        nzhour     = 0
        nzjday     = 100
        nzmin      = 23
        nzmsec     = 465
        nzsec      = 0
        nzyear     = 2001
        stla       = -39.409999...
        stlo       = 175.75
        unused23   = 0
        """
        # https://ds.iris.edu/files/sac-manual/commands/listhdr.html
        print(self._format_header_str(hdrlist))

    def lh(self, *args, **kwargs):
        """Alias of listhdr method."""
        self.listhdr(*args, **kwargs)

    def __str__(self):
        return self._format_header_str()

    def __repr__(self):
        # XXX: run self._flush_headers first?
        # TODO: make this somehow more readable.
        h = sorted(self._header.items())
        fmt = ", {}={!r}" * len(h)
        argstr = fmt.format(*chain.from_iterable(h))[2:]
        return self.__class__.__name__ + "(" + argstr + ")"

    def copy(self):
        return deepcopy(self)

    def _flush_headers(self):
        """
        Flush to the header arrays any header property values that may not be
        reflected there, such as data min/max/mean, npts, e.

        """
        # XXX: do I really care which byte order it is?
        # self.data = np.require(self.data, native_str('<f4'))
        self._hi[HD.INTHDRS.index('npts')] = self.npts
        self._hf[HD.FLOATHDRS.index('e')] = self.e
        self._hf[HD.FLOATHDRS.index('depmin')] = self.depmin
        self._hf[HD.FLOATHDRS.index('depmax')] = self.depmax
        self._hf[HD.FLOATHDRS.index('depmen')] = self.depmen

    def _allt(self, shift):
        """
        Shift all relative time headers by some value (addition).

        Similar to SAC's "chnhdr allt".

        Note
        ----
        This method is triggered by setting an instance's iztype or changing
        its reference time, which is the most likely use case for this
        functionality.  If what you're trying to do is set an origin time and
        make a file origin-based:

        SAC> CHNHDR O GMT 1982 123 13 37 10 103
        SAC>  LISTHDR O
        O 123.103
        SAC>  CHNHDR ALLT -123.103 IZTYPE IO

        ...it is recommended to just make sure your target reference header is
        set and correct, and set the iztype:

        >>> from obspy import UTCDateTime
        >>> from obspy.core.util import get_example_file
        >>> file_ = get_example_file("test.sac")
        >>> sac = SACTrace.read(file_)
        >>> sac.o = UTCDateTime(year=1982, julday=123,
        ...                     hour=13, minute=37,
        ...                     second=10, microsecond=103)
        >>> sac.iztype = 'io'

        The iztype setter will deal with shifting the time values.

        """
        for hdr in ['b', 'o', 'a', 'f'] + ['t' + str(i) for i in range(10)]:
            val = getattr(self, hdr)
            if val is not None:
                setattr(self, hdr, val + shift)

    def _set_distances(self, force=False):
        """
        Calculate dist, az, baz, gcarc.  If force=True, ignore lcalda.
        Raises SacHeaderError if force=True and geographic headers are unset.

        """
        if self.lcalda or force:
            try:
                m, az, baz = gps2dist_azimuth(self.evla, self.evlo, self.stla,
                                              self.stlo)
                dist = m / 1000.0
                gcarc = kilometer2degrees(dist)
                self._hf[HD.FLOATHDRS.index('az')] = az
                self._hf[HD.FLOATHDRS.index('baz')] = baz
                self._hf[HD.FLOATHDRS.index('dist')] = dist
                self._hf[HD.FLOATHDRS.index('gcarc')] = gcarc
            except (ValueError, TypeError):
                # one or more of the geographic values is None
                if force:
                    msg = ("Not enough information to calculate distance, "
                           "azimuth.")
                    raise SacHeaderError(msg)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#  Purpose: Convenience imports for obspy
#   Author: Robert Barsch
#           Moritz Beyreuther
#           Lion Krischer
#           Tobias Megies
#
# Copyright (C) 2008-2014 Robert Barsch, Moritz Beyreuther, Lion Krischer,
#                         Tobias Megies
# -----------------------------------------------------------------------------
"""
ObsPy: A Python Toolbox for seismology/seismological observatories
==================================================================

ObsPy is an open-source project dedicated to provide a Python framework for
processing seismological data. It provides parsers for common file formats
and seismological signal processing routines which allow the manipulation of
seismological time series.

The goal of the ObsPy project is to facilitate rapid application development
for seismology.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import PY2, native_str

import warnings
import requests

# don't change order
from obspy.core.utcdatetime import UTCDateTime  # NOQA
from obspy.core.util import _get_version_string
__version__ = _get_version_string(abbrev=10)
from obspy.core.trace import Trace  # NOQA
from obspy.core.stream import Stream, read
from obspy.core.event import read_events, Catalog
from obspy.core.inventory import read_inventory, Inventory  # NOQA
from obspy.core.util.obspy_types import (  # NOQA
    ObsPyException, ObsPyReadingError)


__all__ = ["UTCDateTime", "Trace", "__version__", "Stream", "read",
           "read_events", "Catalog", "read_inventory", "ObsPyException",
           "ObsPyReadingError"]
__all__ = [native_str(i) for i in __all__]


# insert supported read/write format plugin lists dynamically in docstrings
from obspy.core.util.base import _add_format_plugin_table


_add_format_plugin_table(read, "waveform", "read", numspaces=4)
_add_format_plugin_table(read_events, "event", "read", numspaces=4)
_add_format_plugin_table(read_inventory, "inventory", "read", numspaces=4)
_add_format_plugin_table(Stream.write, "waveform", "write", numspaces=8)
_add_format_plugin_table(Catalog.write, "event", "write", numspaces=8)
_add_format_plugin_table(Inventory.write, "inventory", "write", numspaces=8)


if requests.__version__ in ('2.12.0', '2.12.1', '2.12.2'):
    msg = ("ObsPy has some known issues with 'requests' version {} (see "
           "github issue #1599). Please consider updating module 'requests' "
           "to a newer version.").format(requests.__version__)
    warnings.warn(msg)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

import unittest

from obspy.core.compatibility import py3_round


class CompatibilityTestCase(unittest.TestCase):
    def test_py3_round(self):
        """
        Ensure py3_round rounds correctly and returns expected data types.
        """
        # list of tuples; (input, ndigits, expected, excpected type)
        test_list = [
            (.222, 2, .22, float),
            (1516047903968282880, -3, 1516047903968283000, int),
            (1.499999999, None, 1, int),
            (.0222, None, 0, int),
            (12, -1, 10, int),
            (15, -1, 20, int),
            (15, -2, 0, int),
        ]
        for number, ndigits, expected, expected_type in test_list:
            # this is a fix for py3.4 which cannot take None as ndigits
            if ndigits is not None:
                out = py3_round(number, ndigits)
            else:
                out = py3_round(number)
            self.assertEqual(out, expected)
            self.assertIsInstance(out, expected_type)


def suite():
    return unittest.makeSuite(CompatibilityTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import ctypes as C  # NOQA
import io
import multiprocessing
import os
import platform
import random
import signal
import sys
import unittest
import warnings

import numpy as np

from obspy import Stream, Trace, UTCDateTime, read
from obspy.core.compatibility import from_buffer, mock
from obspy.core.util import NamedTemporaryFile
from obspy.core.util.attribdict import AttribDict
from obspy.core.util.testing import WarningsCapture
from obspy.io.mseed import (InternalMSEEDError, InternalMSEEDWarning,
                            ObsPyMSEEDFilesizeTooSmallError,
                            ObsPyMSEEDFilesizeTooLargeError)
from obspy.io.mseed import util
from obspy.io.mseed.core import _read_mseed, _write_mseed
from obspy.io.mseed.headers import clibmseed
from obspy.io.mseed.msstruct import _MSStruct


# some Python version don't support negative timestamps
NO_NEGATIVE_TIMESTAMPS = False
try:
    UTCDateTime(-50000)
except Exception:
    NO_NEGATIVE_TIMESTAMPS = True


def _test_function(filename):
    """
    Internal function used by MSEEDSpecialIssueTestCase.test_infinite_loop
    """
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("always")
        try:
            st = read(filename)  # noqa @UnusedVariable
        except (ValueError, InternalMSEEDError):
            # Should occur with broken files
            pass


class MSEEDSpecialIssueTestCase(unittest.TestCase):
    """
    """
    def setUp(self):
        # Directory where the test files are located
        self.path = os.path.dirname(__file__)
        # mseed steim compression is big endian
        if sys.byteorder == 'little':
            self.swap = 1
        else:
            self.swap = 0

    def test_invalid_record_length(self):
        """
        An invalid record length should raise an exception.
        """
        npts = 6000
        np.random.seed(815)  # make test reproducible
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            data = np.random.randint(-1000, 1000, npts).astype(np.int32)
            st = Stream([Trace(data=data)])
            # Writing should fail with invalid record lengths.
            # Not a power of 2.
            self.assertRaises(ValueError, _write_mseed, st, tempfile,
                              format="MSEED", reclen=1000)
            # Too small.
            self.assertRaises(ValueError, _write_mseed, st, tempfile,
                              format="MSEED", reclen=8)
            # Not a number.
            self.assertRaises(ValueError, _write_mseed, st, tempfile,
                              format="MSEED", reclen='A')

    def test_invalid_encoding(self):
        """
        An invalid encoding should raise an exception.
        """
        npts = 6000
        np.random.seed(815)  # make test reproducible
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            data = np.random.randint(-1000, 1000, npts).astype(np.int32)
            st = Stream([Trace(data=data)])
            # Writing should fail with invalid record lengths.
            # Wrong number.
            self.assertRaises(ValueError, _write_mseed, st, tempfile,
                              format="MSEED", encoding=2)
            # Wrong Text.
            self.assertRaises(ValueError, _write_mseed, st, tempfile,
                              format="MSEED", encoding='FLOAT_64')

    def test_ctypes_arg_types(self):
        """
        Test that ctypes argtypes are set for type checking
        """
        argument_error = C.ArgumentError
        cl = clibmseed
        args = [C.pointer(C.pointer(C.c_int())), 'a', 1, 1.5, 1, 0, 0, 0, 0]
        self.assertRaises(argument_error, cl.ms_readtraces, *args)
        self.assertRaises(TypeError, cl.ms_readtraces, *args[:-1])
        self.assertRaises(argument_error, cl.ms_readmsr_r, *args)
        self.assertRaises(TypeError, cl.ms_readmsr_r, *args[:-1])
        self.assertRaises(argument_error, cl.mst_printtracelist, *args[:5])
        self.assertRaises(argument_error, cl.ms_detect, *args[:4])
        args.append(1)  # 10 argument function
        self.assertRaises(argument_error, cl.mst_packgroup, *args)
        args = ['hallo']  # one argument functions
        self.assertRaises(argument_error, cl.msr_starttime, *args)
        self.assertRaises(argument_error, cl.msr_endtime, *args)
        self.assertRaises(argument_error, cl.mst_init, *args)
        self.assertRaises(argument_error, cl.mst_free, *args)
        self.assertRaises(argument_error, cl.mst_initgroup, *args)
        self.assertRaises(argument_error, cl.mst_freegroup, *args)
        self.assertRaises(argument_error, cl.msr_init, *args)

    def test_broken_last_record(self):
        """
        Test if Libmseed is able to read files with broken last record. Use
        both methods, readMSTracesViaRecords and readMSTraces
        """
        file = os.path.join(self.path, "data", "brokenlastrecord.mseed")

        # independent reading of the data, 128 Bytes header
        d = np.fromfile(file, dtype=np.uint8)[128:]
        data = util._unpack_steim_2(d, 5980, swapflag=self.swap,
                                    verbose=0)

        # test readMSTraces. Will raise an internal warning.
        with WarningsCapture() as w:
            data_record = _read_mseed(file)[0].data

        # This will raise 18 (!) warnings. It will skip 17 * 128 bytes due
        # to it not being a SEED records and then complain that the remaining
        # 30 bytes are not enough to constitute a full SEED record.
        self.assertEqual(len(w), 18)
        self.assertEqual(w[0].category, InternalMSEEDWarning)

        # Test against reference data.
        self.assertEqual(len(data_record), 5980)
        last10samples = [2862, 2856, 2844, 2843, 2851,
                         2853, 2853, 2854, 2857, 2863]
        np.testing.assert_array_equal(data_record[-10:], last10samples)

        # Also test against independently unpacked data.
        np.testing.assert_allclose(data_record, data)

    def test_one_sample_overlap(self):
        """
        Both methods readMSTraces and readMSTracesViaRecords should recognize a
        single sample overlap.
        """
        # create a stream with one sample overlapping
        trace1 = Trace(data=np.zeros(1000))
        trace2 = Trace(data=np.zeros(10))
        trace2.stats.starttime = UTCDateTime(999)
        st = Stream([trace1, trace2])
        # write into MSEED
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            _write_mseed(st, tempfile, format="MSEED")
            # read it again
            new_stream = _read_mseed(tempfile)
            self.assertEqual(len(new_stream), 2)

    def test_bug_write_read_float32_seed_win32(self):
        """
        Test case for issue #64.
        """
        # create stream object
        data = np.array([395.07809448, 395.0782, 1060.28112793, -1157.37487793,
                         -1236.56237793, 355.07028198, -1181.42175293],
                        dtype=np.float32)
        st = Stream([Trace(data=data)])
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            _write_mseed(st, tempfile, format="MSEED")
            # read temp file directly without libmseed
            with open(tempfile, 'rb') as fp:
                fp.seek(56)
                dtype = np.dtype(native_str('>f4'))
                bin_data = from_buffer(fp.read(7 * dtype.itemsize),
                                       dtype=dtype)
            np.testing.assert_array_equal(data, bin_data)
            # read via ObsPy
            st2 = _read_mseed(tempfile)
        # test results
        np.testing.assert_array_equal(data, st2[0].data)

    @unittest.skipIf(NO_NEGATIVE_TIMESTAMPS,
                     'times before 1970 are not supported on this operation '
                     'system')
    def test_write_with_date_time_before_1970(self):
        """
        Write an stream via libmseed with a datetime before 1970.

        This test depends on the platform specific localtime()/gmtime()
        function.
        """
        # create trace
        tr = Trace(data=np.empty(1000))
        tr.stats.starttime = UTCDateTime("1969-01-01T00:00:00")
        # write file
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            _write_mseed(Stream([tr]), tempfile, format="MSEED")
            # read again
            stream = _read_mseed(tempfile)
            stream.verify()

    def test_invalid_data_type(self):
        """
        Writing data of type int64 and int8 are not supported.

        int64 data can now be written since #2356 if it can be downcast to
        int32.
        """
        npts = 6000
        np.random.seed(815)  # make test reproducible
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            # int64
            data = np.random.randint(-1000, 1000, npts).astype(np.int64)
            # add a data point that can not be downcast (even though we have a
            # separate test for this)
            data[4] = np.iinfo(np.int32).max + 2
            st = Stream([Trace(data=data)])
            self.assertRaises(Exception, st.write, tempfile, format="MSEED")
            # int8
            data = np.random.randint(-1000, 1000, npts).astype(np.int8)
            st = Stream([Trace(data=data)])
            self.assertRaises(Exception, st.write, tempfile, format="MSEED")

    def test_write_wrong_encoding(self):
        """
        Test to write a floating point mseed file with encoding STEIM1.
        An exception should be raised.
        """
        file = os.path.join(self.path, "data",
                            "BW.BGLD.__.EHE.D.2008.001.first_record")
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            # Read the data and convert them to float
            st = read(file)
            st[0].data = st[0].data.astype(np.float32) + .5
            # Type is not consistent float32 cannot be compressed with STEIM1,
            # therefore a exception should be raised.
            self.assertRaises(Exception, st.write, tempfile, format="MSEED",
                              encoding=10)

    def test_write_wrong_encoding_via_mseed_stats(self):
        """
        Test to write a floating point mseed file with encoding STEIM1 with the
        encoding set in stats.mseed.encoding.
        This will just raise a warning.
        """
        file = os.path.join(self.path, "data",
                            "BW.BGLD.__.EHE.D.2008.001.first_record")
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            # Read the data and convert them to float
            st = read(file)
            st[0].data = st[0].data.astype(np.float32) + .5
            # Type is not consistent float32 cannot be compressed with STEIM1,
            # therefore a warning should be raised.
            self.assertEqual(st[0].stats.mseed.encoding, 'STEIM1')
            with warnings.catch_warnings(record=True):
                warnings.simplefilter('error', UserWarning)
                self.assertRaises(UserWarning, st.write, tempfile,
                                  format="MSEED")

    def test_wrong_record_length_as_argument(self):
        """
        Specifying a wrong record length should raise an error.
        """
        file = os.path.join(self.path, 'data', 'encoding',
                            'float32_Float32_bigEndian.mseed')
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            # invalid reclen
            read(file, reclen=111)
            self.assertTrue('Invalid record length' in str(w[0].message))
            self.assertEqual(w[0].category, UserWarning)
            # wrong reclen - raises and also displays a warning
            self.assertRaises(Exception, read, file, reclen=4096)
            self.assertTrue('reclen exceeds buflen' in str(w[1].message))
            self.assertEqual(w[1].category, InternalMSEEDWarning)

    def test_read_with_missing_blockette010(self):
        """
        Reading a Full/Mini-SEED w/o blockette 010 but blockette 008.
        """
        # 1 - Mini-SEED
        file = os.path.join(self.path, 'data', 'blockette008.mseed')
        tr = read(file)[0]
        self.assertEqual('BW.PART..EHZ', tr.id)
        self.assertEqual(1642, tr.stats.npts)
        # 2 - full SEED
        file = os.path.join(self.path, 'data',
                            'RJOB.BW.EHZ.D.300806.0000.fullseed')
        tr = read(file)[0]
        self.assertEqual('BW.RJOB..EHZ', tr.id)
        self.assertEqual(412, tr.stats.npts)

    def test_issue160(self):
        """
        Tests issue #160.

        Reading the header of SEED file.
        """
        file = os.path.join(self.path, 'data',
                            'BW.BGLD.__.EHE.D.2008.001.first_10_records')
        tr_one = read(file)[0]
        tr_two = read(file, headonly=True)[0]
        ms = AttribDict({'record_length': 512, 'encoding': 'STEIM1',
                         'filesize': 5120, 'dataquality': 'D',
                         'number_of_records': 10, 'byteorder': '>'})
        for tr in tr_one, tr_two:
            self.assertEqual('BW.BGLD..EHE', tr.id)
            self.assertEqual(ms, tr.stats.mseed)
            self.assertEqual(4120, tr.stats.npts)
            self.assertEqual(UTCDateTime(2008, 1, 1, 0, 0, 20, 510000),
                             tr.stats.endtime)

    def test_issue217(self):
        """
        Tests issue #217.

        Reading a MiniSEED file without sequence numbers and a record length of
        1024.
        """
        file = os.path.join(self.path, 'data',
                            'reclen_1024_without_sequence_numbers.mseed')
        tr = read(file)[0]
        ms = AttribDict({'record_length': 1024, 'encoding': 'STEIM1',
                         'filesize': 2048, 'dataquality': 'D',
                         'number_of_records': 2, 'byteorder': '>'})
        self.assertEqual('XX.STF1..HHN', tr.id)
        self.assertEqual(ms, tr.stats.mseed)
        self.assertEqual(932, tr.stats.npts)
        self.assertEqual(UTCDateTime(2007, 5, 31, 22, 45, 46, 720000),
                         tr.stats.endtime)

    def test_issue296(self):
        """
        Tests issue #296.
        """
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            # 1 - transform to np.float64 values
            st = read()
            for tr in st:
                tr.data = tr.data.astype(np.float64)
            # write a single trace automatically detecting encoding
            st[0].write(tempfile, format="MSEED")
            # write a single trace automatically detecting encoding
            st.write(tempfile, format="MSEED")
            # write a single trace with encoding 5
            st[0].write(tempfile, format="MSEED", encoding=5)
            # write a single trace with encoding 5
            st.write(tempfile, format="MSEED", encoding=5)
            # 2 - transform to np.float32 values
            st = read()
            for tr in st:
                tr.data = tr.data.astype(np.float32)
            # write a single trace automatically detecting encoding
            st[0].write(tempfile, format="MSEED")
            # write a single trace automatically detecting encoding
            st.write(tempfile, format="MSEED")
            # write a single trace with encoding 4
            st[0].write(tempfile, format="MSEED", encoding=4)
            # write a single trace with encoding 4
            st.write(tempfile, format="MSEED", encoding=4)
            # 3 - transform to np.int32 values
            st = read()
            for tr in st:
                tr.data = tr.data.astype(np.int32)
            # write a single trace automatically detecting encoding
            st[0].write(tempfile, format="MSEED")
            # write a single trace automatically detecting encoding
            st.write(tempfile, format="MSEED")
            # write a single trace with encoding 3
            st[0].write(tempfile, format="MSEED", encoding=3)
            # write the whole stream with encoding 3
            st.write(tempfile, format="MSEED", encoding=3)
            # write a single trace with encoding 10
            st[0].write(tempfile, format="MSEED", encoding=10)
            # write the whole stream with encoding 10
            st.write(tempfile, format="MSEED", encoding=10)
            # write a single trace with encoding 11
            st[0].write(tempfile, format="MSEED", encoding=11)
            # write the whole stream with encoding 11
            st.write(tempfile, format="MSEED", encoding=11)
            # 4 - transform to np.int16 values
            st = read()
            for tr in st:
                tr.data = tr.data.astype(np.int16)
            # write a single trace automatically detecting encoding
            st[0].write(tempfile, format="MSEED")
            # write a single trace automatically detecting encoding
            st.write(tempfile, format="MSEED")
            # write a single trace with encoding 1
            st[0].write(tempfile, format="MSEED", encoding=1)
            # write the whole stream with encoding 1
            st.write(tempfile, format="MSEED", encoding=1)
            # 5 - transform to ASCII values
            st = read()
            for tr in st:
                tr.data = tr.data.astype(native_str('|S1'))
            # write a single trace automatically detecting encoding
            st[0].write(tempfile, format="MSEED")
            # write a single trace automatically detecting encoding
            st.write(tempfile, format="MSEED")
            # write a single trace with encoding 0
            st[0].write(tempfile, format="MSEED", encoding=0)
            # write the whole stream with encoding 0
            st.write(tempfile, format="MSEED", encoding=0)

    def test_issue289(self):
        """
        Tests issue #289.

        Reading MiniSEED using start/end time outside of data should result in
        an empty Stream object.
        """
        # 1
        file = os.path.join(self.path, 'data', 'steim2.mseed')
        st = read(file, starttime=UTCDateTime() - 10, endtime=UTCDateTime())
        self.assertEqual(len(st), 0)
        # 2
        file = os.path.join(self.path, 'data', 'fullseed.mseed')
        st = read(file, starttime=UTCDateTime() - 10, endtime=UTCDateTime())
        self.assertEqual(len(st), 0)

    def test_issue312(self):
        """
        Tests issue #312

        The blkt_link struct was defined wrong.
        """
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')
        # start and end time
        ms = _MSStruct(filename)
        ms.read(-1, 0, 1, 0)
        blkt_link = ms.msr.contents.blkts.contents
        # The first blockette usually begins after 48 bytes. In the test file
        # it does.
        self.assertEqual(blkt_link.blktoffset, 48)
        # The first blockette is blockette 1000 in this file.
        self.assertEqual(blkt_link.blkt_type, 1000)
        # Only one blockette.
        self.assertEqual(blkt_link.next_blkt, 0)
        # Blockette data is 8 bytes - 4 bytes for the blockette header.
        self.assertEqual(blkt_link.blktdatalen, 4)
        del ms

    def test_issue272(self):
        """
        Tests issue #272

        Option headonly should not read the actual waveform data.
        """
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')
        # everything
        st = read(filename)
        self.assertEqual(st[0].stats.npts, 4120)
        self.assertEqual(len(st[0].data), 4120)
        # headers only
        st = read(filename, headonly=True)
        self.assertEqual(st[0].stats.npts, 4120)
        self.assertEqual(len(st[0].data), 0)

    def test_issue325(self):
        """
        Tests issue #325: Use selection with non default dataquality flag.
        """
        filename = os.path.join(self.path, 'data', 'dataquality-m.mseed')
        # 1 - read all
        st = read(filename)
        self.assertEqual(len(st), 3)
        t1 = st[0].stats.starttime
        t2 = st[0].stats.endtime
        # 2 - select full time window
        st2 = read(filename, starttime=t1, endtime=t2)
        self.assertEqual(len(st2), 3)
        for tr in st2:
            del tr.stats.processing
        self.assertEqual(st, st2)
        # 3 - use selection
        st2 = read(filename, starttime=t1, endtime=t2, sourcename='*.*.*.*')
        self.assertEqual(len(st2), 3)
        for tr in st2:
            del tr.stats.processing
        self.assertEqual(st, st2)
        st2 = read(filename, starttime=t1, endtime=t2, sourcename='*')
        self.assertEqual(len(st2), 3)
        for tr in st2:
            del tr.stats.processing
        self.assertEqual(st, st2)
        # 4 - selection without times
        st2 = read(filename, sourcename='*.*.*.*')
        self.assertEqual(len(st2), 3)
        self.assertEqual(st, st2)
        st2 = read(filename, sourcename='*')
        self.assertEqual(len(st2), 3)
        self.assertEqual(st, st2)

    def test_issue332(self):
        """
        Tests issue #332

        Writing traces with wrong encoding in stats should raise only a user
        warning.
        """
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            st = read()
            tr = st[0]
            tr.data = tr.data.astype(np.float64) + .5
            tr.stats.mseed = {'encoding': 0}
            with warnings.catch_warnings(record=True):
                warnings.simplefilter('error', UserWarning)
                self.assertRaises(UserWarning, st.write, tempfile,
                                  format="MSEED")

    def test_issue341(self):
        """
        Tests issue #341

        Read/write of MiniSEED files with huge sampling rates/delta values.
        """
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            # 1 - sampling rate
            st = read()
            tr = st[0]
            tr.stats.sampling_rate = 1000000000.0
            tr.write(tempfile, format="MSEED")
            # read again
            st = read(tempfile)
            self.assertEqual(st[0].stats.sampling_rate, 1000000000.0)
            # 2 - delta
            st = read()
            tr = st[0]
            tr.stats.delta = 10000000.0
            tr.write(tempfile, format="MSEED")
            # read again
            st = read(tempfile)
            self.assertAlmostEqual(st[0].stats.delta, 10000000.0, 0)

    def test_issue485(self):
        """
        Test reading floats and doubles, which are bytswapped nans
        """
        ref = [-1188.07800293, 638.16400146, 395.07809448, 1060.28112793]
        for filename in ('nan_float32.mseed', 'nan_float64.mseed'):
            filename = os.path.join(self.path, 'data', 'encoding', filename)
            data = read(filename)[0].data.tolist()
            np.testing.assert_array_almost_equal(
                data, ref, decimal=8, err_msg='Data of file %s not equal' %
                filename)

    def test_enforcing_reading_byteorder(self):
        """
        Tests if setting the byte order of the header for reading is passed to
        the C functions.

        Quite simple. It just checks if reading with the correct byte order
        works and reading with the wrong byte order fails.
        """
        tr = Trace(data=np.arange(10, dtype=np.int32))

        # Test with little endian.
        memfile = io.BytesIO()
        tr.write(memfile, format="mseed", byteorder="<")
        memfile.seek(0, 0)
        # Reading little endian should work just fine.
        tr2 = read(memfile, header_byteorder="<")[0]
        memfile.seek(0, 0)
        self.assertEqual(tr2.stats.mseed.byteorder, "<")
        # Remove the mseed specific header fields. These are obviously not
        # equal.
        del tr2.stats.mseed
        del tr2.stats._format
        self.assertEqual(tr, tr2)
        # Wrong byte order raises.
        self.assertRaises(ValueError, read, memfile, header_byteorder=">")

        # Same test with big endian
        memfile = io.BytesIO()
        tr.write(memfile, format="mseed", byteorder=">")
        memfile.seek(0, 0)
        # Reading big endian should work just fine.
        tr2 = read(memfile, header_byteorder=">")[0]
        memfile.seek(0, 0)
        self.assertEqual(tr2.stats.mseed.byteorder, ">")
        # Remove the mseed specific header fields. These are obviously not
        # equal.
        del tr2.stats.mseed
        del tr2.stats._format
        self.assertEqual(tr, tr2)
        # Wrong byte order raises.
        self.assertRaises(ValueError, read, memfile, header_byteorder="<")

    def test_long_year_range(self):
        """
        Tests reading and writing years 1900 to 2100.
        """
        tr = Trace(np.arange(5, dtype=np.float32))

        # Year 2056 is non-deterministic for days 1, 256 and 257. These three
        # dates are simply simply not supported right now. See the libmseed
        # documentation for more details.
        # Use every 5th year. Otherwise the test takes too long. Use 1901 as
        # start to get year 2056.
        years = range(1901, 2101, 5)
        for year in years:
            for byteorder in ["<", ">"]:
                memfile = io.BytesIO()
                # Get some random time with the year and the byte order as the
                # seed.
                random.seed(year + ord(byteorder))
                tr.stats.starttime = UTCDateTime(
                    year,
                    julday=random.randrange(1, 365),
                    hour=random.randrange(0, 24),
                    minute=random.randrange(0, 60),
                    second=random.randrange(0, 60))
                if year == 2056:
                    tr.stats.starttime = UTCDateTime(2056, 2, 1)
                tr.write(memfile, format="mseed")
                st2 = read(memfile)
                self.assertEqual(len(st2), 1)
                tr2 = st2[0]
                # Remove the mseed specific header fields. These are obviously
                # not equal.
                del tr2.stats.mseed
                del tr2.stats._format
                self.assertEqual(tr, tr2)

    def test_full_seed_with_non_default_dataquality(self):
        """
        Tests the reading of full SEED files with dataqualities other then D.
        """
        # Test the normal one first.
        filename = os.path.join(self.path, 'data', 'fullseed.mseed')
        st = read(filename)
        self.assertEqual(st[0].stats.mseed.dataquality, "D")

        # Test the others. They should also have identical data.
        filename = os.path.join(self.path, 'data',
                                'fullseed_dataquality_M.mseed')
        st = read(filename)
        data_m = st[0].data
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].stats.mseed.dataquality, "M")

        filename = os.path.join(self.path, 'data',
                                'fullseed_dataquality_R.mseed')
        st = read(filename)
        data_r = st[0].data
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].stats.mseed.dataquality, "R")

        filename = os.path.join(self.path, 'data',
                                'fullseed_dataquality_Q.mseed')
        st = read(filename)
        data_q = st[0].data
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].stats.mseed.dataquality, "Q")

        # Assert that the data is the same.
        np.testing.assert_array_equal(data_m, data_r)
        np.testing.assert_array_equal(data_m, data_q)

    @unittest.skipIf(
        "CONDAFORGE" in os.environ and
        os.environ.get("APPVEYOR", "false").lower() == "true",
        'Test is known to fail when building conda package in Appveyor.')
    def test_infinite_loop(self):
        """
        Tests that libmseed doesn't enter an infinite loop on buggy files.
        """
        filename = os.path.join(self.path, 'data', 'infinite-loop.mseed')

        process = multiprocessing.Process(target=_test_function,
                                          args=(filename, ))
        process.start()
        process.join(60)

        fail = process.is_alive()
        process.terminate()
        if process.is_alive():
            if platform.system() == 'Windows':
                os.kill(process.pid, signal.CTRL_BREAK_EVENT)
            else:
                os.kill(process.pid, signal.SIGKILL)
        self.assertFalse(fail)

    def test_microsecond_accuracy_reading_and_writing_before_1970(self):
        """
        Tests that reading and writing data with microsecond accuracy and
        before 1970 works as expected.
        """
        # Test a couple of timestamps. Positive and negative ones.
        timestamps = [123456.789123, -123456.789123, 1.123400, 1.123412,
                      1.123449, 1.123450, 1.123499, -1.123400, -1.123412,
                      -1.123449, -1.123450, -1.123451, -1.123499]

        for timestamp in timestamps:
            starttime = UTCDateTime(timestamp)
            self.assertEqual(starttime.timestamp, timestamp)

            tr = Trace(data=np.linspace(0, 100, 101))
            tr.stats.starttime = starttime

            with io.BytesIO() as fh:
                tr.write(fh, format="mseed")
                fh.seek(0, 0)
                tr2 = read(fh)[0]

            del tr2.stats.mseed
            del tr2.stats._format

            self.assertEqual(tr2.stats.starttime, starttime)
            self.assertEqual(tr2, tr)

    def test_reading_noise_records(self):
        """
        Tests reading a noise record. See #1495.
        """
        file = os.path.join(self.path, "data",
                            "single_record_plus_noise_record.mseed")
        st = read(file)
        self.assertEqual(len(st), 1)
        tr = st[0]
        self.assertEqual(tr.id, "IM.NV32..BHE")
        self.assertEqual(tr.stats.npts, 277)
        self.assertEqual(tr.stats.sampling_rate, 40.0)

    def test_read_file_with_various_noise_records(self):
        """
        Tests reading a custom made file with noise records.
        """
        # This file has the following layout:
        # 1. 256 byte NOISE record
        # 2. 512 byte normal record - station NV30
        # 3. 128 byte NOISE record
        # 4. 512 byte normal record - station NV31
        # 5. 512 byte NOISE record
        # 6. 512 byte NOISE record
        # 7. 512 byte normal record - station NV32
        # 8. 1024 byte NOISE record
        # 9. 512 byte normal record - station NV33
        file = os.path.join(self.path, "data", "various_noise_records.mseed")
        st = read(file)

        self.assertTrue(len(st), 4)
        self.assertTrue(st[0].stats.station, "NV30")
        self.assertTrue(st[1].stats.station, "NV31")
        self.assertTrue(st[2].stats.station, "NV32")
        self.assertTrue(st[3].stats.station, "NV33")

        # Data is the same across all records.
        np.testing.assert_allclose(st[0].data, st[1].data)
        np.testing.assert_allclose(st[0].data, st[2].data)
        np.testing.assert_allclose(st[0].data, st[3].data)

    def test_mseed_zero_data_offset(self):
        """
        Tests that a data offset of zero in the fixed header does not
        confuse ObsPy.

        The file contains three records: a normal one, followed by one with
        a data-offset of zero, followed by another normal one.

        This currently results in three returned traces:

        * CH.PANIX..LHZ |
          2016-08-21T01:41:19.000000Z - 2016-08-21T01:45:30.000000Z |
          1.0 Hz, 252 samples
        * CH.PANIX..LHZ |
          2016-08-21T01:43:37.000000Z - 2016-08-21T01:43:37.000000Z |
          1.0 Hz, 0 samples
        * CH.PANIX..LHZ |
          2016-08-21T01:45:31.000000Z - 2016-08-21T01:49:52.000000Z |
          1.0 Hz, 262 samples
        """
        file = os.path.join(self.path, "data", "bizarre",
                            "mseed_data_offset_0.mseed")
        st = read(file)

        self.assertEqual(len(st), 3)

        tr = st[0]
        self.assertEqual(tr.id, "CH.PANIX..LHZ")
        self.assertEqual(tr.stats.starttime,
                         UTCDateTime("2016-08-21T01:41:19.000000Z"))
        self.assertEqual(tr.stats.endtime,
                         UTCDateTime("2016-08-21T01:45:30.000000Z"))
        self.assertEqual(tr.stats.npts, len(tr.data))
        self.assertEqual(tr.stats.npts, 252)

        tr = st[1]
        self.assertEqual(tr.id, "CH.PANIX..LHZ")
        self.assertEqual(tr.stats.starttime,
                         UTCDateTime("2016-08-21T01:43:37.000000Z"))
        self.assertEqual(tr.stats.endtime,
                         UTCDateTime("2016-08-21T01:43:37.000000Z"))
        self.assertEqual(tr.stats.npts, len(tr.data))
        self.assertEqual(tr.stats.npts, 0)

        tr = st[2]
        self.assertEqual(tr.id, "CH.PANIX..LHZ")
        self.assertEqual(tr.stats.starttime,
                         UTCDateTime("2016-08-21T01:45:31.000000Z"))
        self.assertEqual(tr.stats.endtime,
                         UTCDateTime("2016-08-21T01:49:52.000000Z"))
        self.assertEqual(tr.stats.npts, len(tr.data))
        self.assertEqual(tr.stats.npts, 262)

    def test_mseed_zero_data_headonly(self):
        """
        Tests that records with no data correctly work in headonly mode.
        """
        file = os.path.join(self.path, "data",
                            "three_records_zero_data_in_middle.mseed")

        expected = [
            ("BW.BGLD..EHE", UTCDateTime("2007-12-31T23:59:59.765000Z"),
             UTCDateTime("2008-01-01T00:00:01.820000Z"), 200.0, 412),
            ("BW.BGLD..EHE", UTCDateTime("2008-01-01T00:00:01.825000Z"),
             UTCDateTime("2008-01-01T00:00:01.825000Z"), 200.0, 0),
            ("BW.BGLD..EHE", UTCDateTime("2008-01-01T00:00:03.885000Z"),
             UTCDateTime("2008-01-01T00:00:05.940000Z"), 200.0, 412)]

        # Default full read.
        st = read(file)
        self.assertEqual(len(st), 3)
        for tr, exp in zip(st, expected):
            self.assertEqual(tr.id, exp[0])
            self.assertEqual(tr.stats.starttime, exp[1])
            self.assertEqual(tr.stats.endtime, exp[2])
            self.assertEqual(tr.stats.sampling_rate, exp[3])
            self.assertEqual(tr.stats.npts, exp[4])

        # Headonly read.
        st = read(file, headonly=True)
        self.assertEqual(len(st), 3)
        for tr, exp in zip(st, expected):
            self.assertEqual(tr.id, exp[0])
            self.assertEqual(tr.stats.starttime, exp[1])
            self.assertEqual(tr.stats.endtime, exp[2])
            self.assertEqual(tr.stats.sampling_rate, exp[3])
            self.assertEqual(tr.stats.npts, exp[4])

    def test_read_file_with_microsecond_wrap(self):
        """
        This is not strictly valid but I encountered such a file in practice
        so I guess it happens. Libmseed can also correctly deal with it.

        The test file is a single record with the .0001 seconds field set to
        10000. SEED strictly allows only 0-9999 in this field.
        """
        file = os.path.join(self.path, "data",
                            "microsecond_wrap.mseed")

        with warnings.catch_warnings(record=True) as w_1:
            warnings.simplefilter("always")
            info = util.get_record_information(file)

        self.assertEqual(w_1[0].message.args[0],
                         'Record contains a fractional seconds (.0001 secs) '
                         'of 10000 - the maximum strictly allowed value is '
                         '9999. It will be interpreted as one or more '
                         'additional seconds.')

        with warnings.catch_warnings(record=True) as w_2:
            warnings.simplefilter("always")
            tr = read(file)[0]

        # First warning is identical.
        self.assertEqual(w_1[0].message.args[0], w_2[0].message.args[0])
        # Second warning is raised by libmseed.
        self.assertEqual(w_2[1].message.args[0],
                         'readMSEEDBuffer(): Record with offset=0 has a '
                         'fractional second (.0001 seconds) of 10000. This '
                         'is not strictly valid but will be interpreted as '
                         'one or more additional seconds.')

        # Make sure libmseed and the internal ObsPy record parser produce
        # the same result.
        self.assertEqual(info["starttime"], tr.stats.starttime)
        self.assertEqual(info["endtime"], tr.stats.endtime)

        # Read with a hex-editor.
        ref_time = UTCDateTime(year=2008, julday=8, hour=4, minute=58,
                               second=5 + 1)
        self.assertEqual(ref_time, info["starttime"])
        self.assertEqual(ref_time, tr.stats.starttime)

    def test_reading_miniseed_with_no_blockette_1000(self):
        """
        Blockette 1000 was only introduced with SEED version 2.3.
        """
        file = os.path.join(self.path, "data", "bizarre",
                            "mseed_no_blkt_1000.mseed")
        st = read(file)

        self.assertEqual(len(st), 1)
        tr = st[0]
        self.assertEqual(tr.stats.npts, 7536)
        np.testing.assert_allclose(tr.stats.sampling_rate, 20.0)
        self.assertEqual(tr.stats.starttime,
                         UTCDateTime(1976, 3, 10, 3, 28))
        self.assertEqual(tr.id, ".GRA1..BHZ")

        # Also test the data to make sure the unpacking was successful.
        np.testing.assert_equal(
            st[0].data[:10],
            [-185, -200, -209, -220, -228, -246, -252, -262, -262, -269])

    def test_writing_of_blockette_100(self):
        """
        Tests the writing of blockette 100 if necessary.
        """
        def has_blkt_100(filename):
            ms = _MSStruct(filename)
            ms.read(-1, 0, 1, 0)
            blkt_link = ms.msr.contents.blkts
            has_blkt_100 = False
            while True:
                try:
                    if blkt_link.contents.blkt_type == 100:
                        has_blkt_100 = True
                        break
                    blkt_link = blkt_link.contents.next
                except ValueError:
                    break
            del ms
            return has_blkt_100

        tr = read()[0]
        tr.data = tr.data[:10]

        with NamedTemporaryFile() as tf:
            tempfile = tf.name

            # A "clean" sampling rate does not require blockette 100.
            sr = 200.0
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

            # A more detailed one does.
            sr = 199.9997
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertTrue(has_blkt_100(tempfile))

            # A very large (but "clean") one does not need it.
            sr = 1E6
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

            # Same for a very small but "clean" one.
            sr = 1E-6
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

            # But small perturbations resulting in "unclean" sampling rates
            # will cause blockette 100 to be written.
            sr = 1E6 + 0.123456
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertTrue(has_blkt_100(tempfile))
            sr = 1E-6 + 0.325247 * 1E-7
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertTrue(has_blkt_100(tempfile))

            # Three more "clean" ones.
            sr = 1.0
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

            sr = 0.5
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

            sr = 0.1
            tr.stats.sampling_rate = sr
            tr.write(tempfile, format="mseed")
            self.assertEqual(
                np.float32(read(tempfile)[0].stats.sampling_rate),
                np.float32(sr))
            self.assertFalse(has_blkt_100(tempfile))

    def test_reading_file_with_data_offset_of_48(self):
        """
        Tests reading a file which has a data offset of 48 bytes.

        It thus does not have a single blockette.
        """
        file = os.path.join(
            self.path, "data",
            "mseed_not_a_single_blkt_48byte_data_offset.mseed")
        st = read(file)

        self.assertEqual(len(st), 1)
        tr = st[0]
        self.assertEqual(tr.stats.npts, 3632)
        np.testing.assert_allclose(tr.stats.sampling_rate, 20.0)
        self.assertEqual(tr.stats.starttime,
                         UTCDateTime(1995, 9, 22, 0, 0, 18, 238400))
        self.assertEqual(tr.id, "XX.TEST..BHE")

        np.testing.assert_equal(
            st[0].data[:6],
            [337, 396, 454, 503, 547, 581])

    def test_reading_truncated_miniseed_files(self):
        """
        Regression test to guard against a segfault.
        """
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')

        with io.open(filename, 'rb') as fh:
            data = fh.read()

        data = data[:-257]
        # This is the offset for the record that later has to be recorded in
        # the warning.
        self.assertEqual(len(data) - 255, 4608)

        # The file now lacks information at the end. This will read the file
        # until that point and raise a warning that some things are missing.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with io.BytesIO(data) as buf:
                st = _read_mseed(buf)
        self.assertEqual(len(st), 1)
        self.assertEqual(len(w), 1)
        self.assertIs(w[0].category, InternalMSEEDWarning)
        self.assertEqual("readMSEEDBuffer(): Unexpected end of file when "
                         "parsing record starting at offset 4608. The rest of "
                         "the file will not be read.", w[0].message.args[0])

    def test_reading_truncated_miniseed_files_case_2(self):
        """
        Second test in the same vain as
        test_reading_truncated_miniseed_files. Previously forgot a `<=` test.
        """
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')

        with io.open(filename, 'rb') as fh:
            data = fh.read()

        data = data[:-256]
        # This is the offset for the record that later has to be recorded in
        # the warning.
        self.assertEqual(len(data) - 256, 4608)

        # The file now lacks information at the end. This will read the file
        # until that point and raise a warning that some things are missing.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with io.BytesIO(data) as buf:
                st = _read_mseed(buf)
        self.assertEqual(len(st), 1)
        self.assertEqual(len(w), 1)
        self.assertIs(w[0].category, InternalMSEEDWarning)
        self.assertEqual("readMSEEDBuffer(): Unexpected end of file when "
                         "parsing record starting at offset 4608. The rest of "
                         "the file will not be read.", w[0].message.args[0])

    def test_reading_less_than_128_bytes(self):
        """
        128 bytes is the smallest possible MiniSEED record.

        Reading anything smaller should result in an error.
        """
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')

        with io.open(filename, 'rb') as fh:
            data = fh.read()

        # Reading at exactly 128 bytes offset will result in a truncation
        # warning.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            with io.BytesIO(data[:128]) as buf:
                st = _read_mseed(buf)
        self.assertEqual(len(st), 0)  # nothing is read here.
        self.assertGreaterEqual(len(w), 1)
        self.assertIs(w[-1].category, InternalMSEEDWarning)
        self.assertEqual("readMSEEDBuffer(): Unexpected end of file when "
                         "parsing record starting at offset 0. The rest of "
                         "the file will not be read.", w[-1].message.args[0])

        # Reading anything less result in an exception.
        with self.assertRaises(ObsPyMSEEDFilesizeTooSmallError) as e:
            with io.BytesIO(data[:127]) as buf:
                _read_mseed(buf)
        self.assertEqual(
            e.exception.args[0],
            "The smallest possible mini-SEED record is made up of 128 bytes. "
            "The passed buffer or file contains only 127.")

    @mock.patch("os.path.getsize")
    def test_reading_file_larger_than_2048_mib(self, getsize_mock):
        """
        ObsPy can currently not directly read files that are larger than
        2^31 bytes. This raises an exception with a description of how to
        get around it.
        """
        getsize_mock.return_value = 2 ** 31 + 1
        filename = os.path.join(self.path, 'data',
                                'BW.BGLD.__.EHE.D.2008.001.first_10_records')
        with self.assertRaises(ObsPyMSEEDFilesizeTooLargeError) as e:
            _read_mseed(filename)
        self.assertEqual(
            e.exception.args[0],
            "ObsPy can currently not directly read mini-SEED files that are "
            "larger than 2^31 bytes (2048 MiB). To still read it, please "
            "read the file in chunks as documented here: "
            "https://github.com/obspy/obspy/pull/1419#issuecomment-221582369")

    def test_read_file_with_non_valid_blocks_in_between(self):
        """
        Test reading MiniSEED files that have some non-valid blocks in-between.
        """
        # This file has two 4096 bytes records.
        filename = os.path.join(self.path, 'data', 'test.mseed')
        with io.open(filename, "rb") as fh:
            rec1 = fh.read(4096)
            rec2 = fh.read(4096)

        reference = _read_mseed(filename)
        del reference[0].stats.mseed

        # Fill with zero bytes.
        for length in (128, 256, 512, 1024, 2048, 4096, 8192):
            with io.BytesIO() as buf:
                buf.write(rec1)
                buf.write(b'\x00' * length)
                buf.write(rec2)
                buf.seek(0, 0)
                # This will raise 1 warning per 128 bytes.
                with WarningsCapture() as w:
                    st = _read_mseed(buf)
                self.assertEqual(len(w), length // 128)

            # Also explicitly test the first warning message which should be
            # identical for all cases.
            self.assertEqual(
                w[0].message.args[0],
                "readMSEEDBuffer(): Not a SEED record. Will skip bytes "
                "4096 to 4223.")

            # Should always be two records.
            self.assertEqual(st[0].stats.mseed.number_of_records, 2)

            # Remove things like file-size and what not.
            del st[0].stats.mseed
            self.assertEqual(reference, st)

        # Try the same thing but fill with random bytes.
        # The seed is not really needed but hopefully guards against the
        # very very rare case of random bytes making up a valid SEED record.
        np.random.seed(34980)
        for length in (128, 256, 512, 1024, 2048, 4096, 8192):
            with io.BytesIO() as buf:
                buf.write(rec1)
                buf.write(np.random.bytes(length))
                buf.write(rec2)
                buf.seek(0, 0)
                # This will raise 1 warning per 128 bytes.
                with WarningsCapture() as w:
                    st = _read_mseed(buf)
                self.assertEqual(len(w), length // 128)

            # Should always be two records.
            self.assertEqual(st[0].stats.mseed.number_of_records, 2)

            # Remove things like file-size and what not.
            del st[0].stats.mseed
            self.assertEqual(reference, st)

    def test_reading_files_with_non_ascii_headers(self):
        """
        Some dataloggers appear to do this.

        See #2177.
        """
        filename = os.path.join(self.path, "data",
                                "gecko_non_ascii_header.ms")
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            st = read(filename)
        # Two identical messages will be raised (but in most common cases only
        # the first will be shown).
        self.assertEqual(len(w), 2)
        self.assertTrue(w[0].message.args[0].startswith(
            "Failed to decode location code as ASCII."))
        self.assertEqual(len(st), 1)
        self.assertEqual(st[0].id, '.GECKO.A.CNZ')


def suite():
    return unittest.makeSuite(MSEEDSpecialIssueTestCase, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# -*- coding: utf-8 -*-
"""
Optional feature generators for ObsPy Trace objects.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA

from obspy.core.util import score_at_percentile


class BandpassPreviewFeature(object):
    """
    Bandpass filter (freqmin=0.1, freqmax=20.0) all trace previews.
    """

    def process(self, trace):
        """
        Bandpass filter (freqmin=0.1, freqmax=20.0) all trace previews.
        """
        # applying bandpass on trace directly - this will not modify the
        # original waveform file but it will influence the preview trace
        trace.filter("bandpass", freqmin=0.1, freqmax=20.0)
        return {}


class MinMaxAmplitudeFeature(object):
    """
    Generates statistics about the amplitude values.
    """

    def process(self, trace):
        """
        Generates statistics about the amplitude values.

        This may take a while to calculate - use a moderate looping interval.

        .. rubric:: Example

        >>> from obspy import Trace
        >>> import numpy as np
        >>> tr = Trace(data=np.arange(-5,5))
        >>> result = MinMaxAmplitudeFeature().process(tr)
        >>> result['max']
        4.0
        >>> result['upper_quantile']
        1.75
        """
        result = {}
        result['min'] = float(trace.data.min())
        result['max'] = float(trace.data.max())
        result['avg'] = float(trace.data.mean())
        result['median'] = float(score_at_percentile(trace.data, 50, False))
        result['lower_quantile'] = float(score_at_percentile(trace.data, 25))
        result['upper_quantile'] = float(score_at_percentile(trace.data, 75))
        return result


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from future.builtins import *  # NOQA
from future.utils import native_str

import csv
import os


class FlinnEngdahl(object):
    """
    Load data from asc files and allow to resolve coordinates or region numbers
    to Flinn Engdahl region names.

    >>> fe = FlinnEngdahl()
    >>> print(fe.get_region(12, 48))
    GERMANY
    >>> print(fe.get_region_by_number(543))
    GERMANY
    """

    data_directory = os.path.join(os.path.dirname(__file__), 'data')

    names_file = os.path.join(data_directory, 'names.asc')
    quadsindex_file = os.path.join(data_directory, 'quadsidx.asc')
    sect_files = (
        os.path.join(data_directory, 'nesect.asc'),
        os.path.join(data_directory, 'nwsect.asc'),
        os.path.join(data_directory, 'sesect.asc'),
        os.path.join(data_directory, 'swsect.asc')
    )
    numbers_file = os.path.join(data_directory, 'Flinn-Engdahl.csv')
    quads_order = ('ne', 'nw', 'se', 'sw')

    def __init__(self):
        self.quads_index = []

        with open(self.names_file, 'r') as fh:
            self.names = [name.strip() for name in fh]

        with open(self.quadsindex_file, 'r') as fh:
            indexes = []
            for index in fh:
                indexes += [n.strip() for n in index.split(' ') if n != '']

        self.lons_per_lat = dict(zip(
            self.quads_order,
            [indexes[x:x + 91] for x in range(0, len(indexes), 91)]
        ))

        self.lat_begins = {}

        for quad, index in self.lons_per_lat.items():
            begin = 0
            end = -1
            begins = []
            n = 0

            for item in index:
                n += 1
                begin = end + 1
                begins.append(begin)
                end += int(item)

            self.lat_begins[quad] = begins

        self.lons = {}
        self.fenums = {}
        for quad, sect_file in zip(self.quads_order, self.sect_files):
            sect = []
            with open(sect_file, 'r') as fh:
                for line in fh:
                    sect += [int(v) for v in line.strip().split(' ')
                             if v != '']

            lons = []
            fenums = []
            n = 0
            for item in sect:
                n += 1
                if n % 2:
                    lons.append(item)
                else:
                    fenums.append(item)

            self.lons[quad] = lons
            self.fenums[quad] = fenums

        with open(self.numbers_file, 'rt') as csvfile:
            fe_csv = csv.reader(csvfile, delimiter=native_str(';'),
                                quotechar=native_str('#'),
                                skipinitialspace=True)
            self.by_number = \
                {int(row[0]): row[1] for row in fe_csv if len(row) > 1}

    def get_quadrant(self, longitude, latitude):
        """
        Return quadrant from given coordinate

        :param longitude: WGS84 longitude
        :type longitude: int or float
        :param latitude: WGS84 latitude
        :type latitude: int or float
        :rtype: string
        :return: Quadrant name (ne, nw, se and sw)
        """
        if longitude >= 0 and latitude >= 0:
            return 'ne'
        if longitude < 0 and latitude >= 0:
            return 'nw'
        if longitude >= 0 and latitude < 0:
            return 'se'
        if longitude < 0 and latitude < 0:
            return 'sw'

    def get_region(self, longitude, latitude):
        """
        Return region from given coordinate

        >>> fe = FlinnEngdahl()
        >>> print(fe.get_region(12, 48))
        GERMANY
        >>> print(fe.get_region(-60, -30))
        NORTHEASTERN ARGENTINA

        :param longitude: WGS84 longitude
        :type longitude: int or float
        :param latitude: WGS84 latitude
        :type latitude: int or float
        :rtype: string
        :return: Flinn Engdahl region name
        """

        if longitude < -180 or longitude > 180:
            raise ValueError
        if latitude < -90 or latitude > 90:
            raise ValueError

        if longitude == -180:
            longitude = 180

        quad = self.get_quadrant(longitude, latitude)

        abs_longitude = int(abs(longitude))
        abs_latitude = int(abs(latitude))

        begin = self.lat_begins[quad][abs_latitude]
        num = int(self.lons_per_lat[quad][abs_latitude])

        my_lons = self.lons[quad][begin:begin + num]
        my_fenums = self.fenums[quad][begin:begin + num]

        n = 0
        for longitude in my_lons:
            if longitude > abs_longitude:
                break
            n += 1

        fe_index = n - 1
        fe_num = my_fenums[fe_index]
        fe_name = self.names[fe_num - 1]

        return fe_name

    def get_region_by_number(self, number):
        """
        Return region with given number

        >>> fe = FlinnEngdahl()
        >>> print(fe.get_region_by_number(123))
        NORTHERN CHILE
        >>> print(fe.get_region_by_number(456))
        MONTANA

        :param number: Region ID
        :type number: int
        :rtype: string
        :return: Flinn Engdahl region name
        """
        return self.by_number[number]


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)
